import logging
import uuid
from typing import Dict, Optional

from ruamel.yaml import YAML
from soda.common.file_system import FileSystemSingleton
from soda.common.yaml_helper import to_yaml_str

logger = logging.getLogger(__name__)
yaml = YAML()


class ConfigHelper:
    """Helper class for handling global Soda config."""

    DEFAULT_CONFIG = {"send_anonymous_usage_stats": True, "user_cookie_id": str(uuid.uuid4())}
    # The possible load paths for the config file - /tmp in the worst case.
    LOAD_PATHS = ["~/.soda/config.yml", ".soda/config.yml", "/tmp/.soda/config.yml"]
    __instance = None
    __config: Dict = {}
    file_system = FileSystemSingleton.INSTANCE

    @staticmethod
    def get_instance(path: Optional[str] = None):
        if ConfigHelper.__instance is None:
            ConfigHelper(path)
        return ConfigHelper.__instance

    def __init__(self, path: Optional[str] = None):
        if ConfigHelper.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            ConfigHelper.__instance = self

        if path:
            self.LOAD_PATHS.insert(0, path)

        self.__config = self.config
        self._config_path = self.LOAD_PATHS[0]
        if not self.__config:
            self.init_config_file()

        self.__ensure_basic_config()

    @property
    def config_path(self) -> str:
        return self._config_path

    @config_path.setter
    def config_path(self, value):
        self._config_path = value

    @property
    def config(self) -> Dict:
        if not self.__config:
            self.__config = self.load_config()

        return self.__config

    def load_config(self) -> Dict:
        config = {}
        for path in self.LOAD_PATHS:
            logger.debug(f"Trying to load Soda Config file {path}.")

            if self.file_system.exists(path):
                content = yaml.load(self.file_system.file_read_as_str(path))
                if content:
                    config = content
                break

        return config

    def get_value(self, key: str, default_value=None):
        """Get value from loaded config."""
        return self.config.get(key, default_value)

    def init_config_file(self) -> None:
        """Init default config file if not present."""

        if self.file_system.exists(self.config_path):
            logger.debug(f"Config file {self.config_path} already exists")
        else:
            for destination in self.LOAD_PATHS:
                try:
                    logger.info(f"Trying to create config YAML file {destination} ...")
                    self.file_system.mkdirs(self.file_system.dirname(destination))
                    self.upsert_config_file(self.DEFAULT_CONFIG)
                    self.config_path = destination
                    return
                except OSError:
                    logger.warning(f"Unable to create config path in {destination}")

    def upsert_value(self, key: str, value: str):
        """Update or insert a value in the config file and refresh loaded state."""
        config = self.config
        config[key] = value
        self.upsert_config_file(config)
        self.__config = self.load_config()

    def upsert_config_file(self, config: Dict):
        """Write provided config into a yaml file."""
        self.file_system.file_write_from_str(
            self.config_path,
            to_yaml_str(config),
        )

    @staticmethod
    def generate_user_cookie_id() -> str:
        return str(uuid.uuid4())

    def __ensure_basic_config(self) -> None:
        for key, value in self.DEFAULT_CONFIG.items():
            if key not in self.config:
                self.upsert_value(key, value)

    @property
    def send_anonymous_usage_stats(self) -> bool:
        return self.config.get("send_anonymous_usage_stats", True)



from __future__ import annotations

import logging
import re

from soda.common.logs import Logs
from soda.common.parser import Parser
from soda.configuration.configuration import Configuration
from soda.sampler.soda_cloud_sampler import SodaCloudSampler
from soda.soda_cloud.soda_cloud import SodaCloud

logger = logging.getLogger(__name__)

DATA_SOURCE = "data_source"
CONNECTION = "connection"


class ConfigurationParser(Parser):
    def __init__(self, configuration: Configuration, logs: Logs, file_path: str):
        super().__init__(file_path=file_path, logs=logs)
        from soda.configuration.configuration import Configuration

        self.configuration: Configuration = configuration

    def parse_environment_yaml_str(self, environment_yaml_str: str):
        environment_yaml_str = self._resolve_jinja(environment_yaml_str)
        environment_dict = self._parse_yaml_str(environment_yaml_str)
        if environment_dict is not None:
            self.__parse_headers(environment_dict)

    def __parse_headers(self, environment_dict: dict) -> None:
        if not environment_dict:
            return

        for environment_header, header_value in environment_dict.items():
            if environment_header.startswith(f"{DATA_SOURCE} "):
                self._push_path_element(environment_header, header_value)
                data_source_name = environment_header[len(f"{DATA_SOURCE} ") :].strip()
                if not re.compile(r"^[a-z_][a-z_0-9]+$").match(data_source_name):
                    self.logs.error(
                        "Invalid data_source name. DataSource names must start with a lower case char or "
                        "an underscore [a-z_], followed by any number of lower case chars, digits or "
                        "underscore [a-z0-9_]"
                    )
                self.configuration.data_source_properties_by_name[data_source_name] = header_value

                self._get_required("type", str)

                data_source_connection = header_value.get("connection")
                if data_source_connection is None:
                    self.logs.error("connection is required", location=self.location)
                elif isinstance(data_source_connection, dict):
                    connection_name = data_source_name
                    self.configuration.connection_properties_by_name[connection_name] = data_source_connection
                    header_value["connection"] = connection_name
                elif not isinstance(data_source_connection, str):
                    self.logs.error(
                        "connection must be a string or a dict",
                        location=self.location,
                    )
                self._pop_path_element()
            elif environment_header.startswith(f"{CONNECTION} "):
                self._push_path_element(environment_header, header_value)
                connection_name = environment_header[len(f"{CONNECTION} ") :].strip()
                self.configuration.connection_properties_by_name[connection_name] = header_value
                self._pop_path_element()

            elif environment_header == "soda_cloud":
                self._push_path_element("soda_cloud", header_value)
                self.configuration.soda_cloud = self.parse_soda_cloud_cfg(header_value)
                if self.configuration.soda_cloud and not header_value.get("disable_samples"):
                    self.configuration.sampler = SodaCloudSampler()
                self._pop_path_element()

            else:
                self.logs.error(
                    f'Invalid configuration header: expected either "{DATA_SOURCE} {{data source name}}" '
                    f'or "{CONNECTION} {{connection name}}"',
                    location=self.location,
                )

    def parse_soda_cloud_cfg(self, soda_cloud_dict: dict):
        api_key = soda_cloud_dict.get("api_key_id")
        api_secret = soda_cloud_dict.get("api_key_secret")
        host = None
        if "host" in soda_cloud_dict:
            host = soda_cloud_dict.get("host")
        port = None
        if "port" in soda_cloud_dict:
            port = soda_cloud_dict.get("port")
        return SodaCloud(
            api_key_id=api_key, api_key_secret=api_secret, host=host, token=None, port=port, logs=self.logs
        )



import sys
import traceback
from typing import Optional


def get_exception_stacktrace(exception) -> Optional[str]:
    if isinstance(exception, BaseException):
        if sys.version_info < (3, 10):
            return "".join(
                traceback.format_exception(etype=type(exception), value=exception, tb=exception.__traceback__)
            )
        return "".join(traceback.format_exception(exception))
    return None



from typing import List

from soda.sodacl.location import Location


class DiscoverTablesCfg:
    def __init__(self, data_source_name: str, location: Location):
        self.data_source_name: str = data_source_name
        self.location: Location = location
        self.include_tables: List[str] = []
        self.exclude_tables: List[str] = []



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from soda.common.logs import Logs
from soda.sampler.sample import Sample


@dataclass
class SampleContext:

    sample: Sample
    sample_name: str
    query: str
    data_source: "DataSource"
    partition: Optional["Partition"]
    column: Optional["Column"]
    scan: "Scan"
    logs: "Logs"

    def get_scan_folder_name(self):
        parts = [
            self.scan._scan_definition_name,
            self.scan._data_timestamp.strftime("%Y%m%d%H%M%S"),
            datetime.now().strftime("%Y%m%d%H%M%S"),
        ]
        return "_".join([part for part in parts if part])

    def get_sample_file_name(self):
        parts = [
            self.partition.table.table_name if self.partition else None,
            self.partition.partition_name if self.partition else None,
            self.column.column_name if self.column else None,
            self.sample_name,
        ]
        return "_".join([part for part in parts if part])



from numbers import Number
from typing import Dict, Optional

from soda.execution.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.column import Column
from soda.execution.data_source_scan import DataSourceScan
from soda.execution.metric import Metric
from soda.execution.partition import Partition
from soda.execution.query import Query
from soda.sodacl.distribution_check_cfg import DistributionCheckCfg

from soda.scientific.distribution.comparison import DistributionChecker


class DistributionCheck(Check):
    def __init__(
        self,
        check_cfg: DistributionCheckCfg,
        data_source_scan: DataSourceScan,
        partition: Optional[Partition] = None,
        column: Optional[Column] = None,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
            name="distribution_difference",
        )

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        distribution_check_cfg: DistributionCheckCfg = self.check_cfg

        sql = self.sql_column_values_query(distribution_check_cfg)

        query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name="get_values_for_distro_check",
            sql=sql,
        )
        query.execute()
        if query.exception is None and query.rows is not None:
            test_data = [row[0] for row in query.rows]

            _, p_value = DistributionChecker(distribution_check_cfg, test_data).run()
            self.check_value = p_value

            self.set_outcome_based_on_check_value()

    def set_outcome_based_on_check_value(self):
        from soda.sodacl.distribution_check_cfg import DistributionCheckCfg

        distribution_check_cfg: DistributionCheckCfg = self.check_cfg
        if self.check_value is not None and (
            distribution_check_cfg.warn_threshold_cfg or distribution_check_cfg.fail_threshold_cfg
        ):
            if isinstance(self.check_value, Number):
                if distribution_check_cfg.fail_threshold_cfg and distribution_check_cfg.fail_threshold_cfg.is_bad(
                    self.check_value
                ):
                    self.outcome = CheckOutcome.FAIL
                elif distribution_check_cfg.warn_threshold_cfg and distribution_check_cfg.warn_threshold_cfg.is_bad(
                    self.check_value
                ):
                    self.outcome = CheckOutcome.WARN
                else:
                    self.outcome = CheckOutcome.PASS
            else:
                self.logs.error(
                    f"Cannot evaluate check: Expected a numeric value, but was {self.check_value}",
                    location=self.check_cfg.location,
                )

    def get_cloud_diagnostics_dict(self) -> dict:
        cloud_diagnostics = super().get_cloud_diagnostics_dict()
        # if self.historic_diff_values:
        #     cloud_diagnostics["historic_diff_values"] = self.historic_diff_values
        return cloud_diagnostics

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = super().get_log_diagnostic_dict()
        # if self.historic_diff_values:
        #     log_diagnostics.update(self.historic_diff_values)
        return log_diagnostics

    def sql_column_values_query(self, distribution_check_cfg):
        column_name = distribution_check_cfg.column_name
        sql = f"SELECT \n" f"  {column_name} \n" f"FROM {self.partition.table.prefixed_table_name}"

        partition_filter = self.partition.sql_partition_filter
        if partition_filter:
            scan = self.data_source_scan.scan
            resolved_filter = scan._jinja_resolve(definition=partition_filter)
            sql += f"\nWHERE {resolved_filter}"

        return sql



from typing import Dict

from soda.execution.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric import Metric
from soda.execution.numeric_query_metric import NumericQueryMetric
from soda.execution.partition import Partition


class RowCountComparisonCheck(Check):
    def __init__(
        self,
        check_cfg: "RowCountComparisonCheckCfg",
        data_source_scan: "DataSourceScan",
        partition: Partition,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
            name="row-count-comparison",
        )
        from soda.execution.table import Table
        from soda.sodacl.row_count_comparison_check_cfg import (
            RowCountComparisonCheckCfg,
        )

        data_source_scan = self.data_source_scan
        scan = data_source_scan.scan

        row_count_comparison_check_cfg: RowCountComparisonCheckCfg = self.check_cfg

        if row_count_comparison_check_cfg.other_data_source_name:
            data_source_scan = scan._get_or_create_data_source_scan(
                row_count_comparison_check_cfg.other_data_source_name
            )

        other_table: Table = data_source_scan.get_or_create_table(row_count_comparison_check_cfg.other_table_name)
        self.other_partition = other_table.get_or_create_partition(row_count_comparison_check_cfg.other_partition_name)

        self.metrics["row_count"] = self.data_source_scan.resolve_metric(
            NumericQueryMetric(
                data_source_scan=self.data_source_scan,
                partition=partition,
                column=None,
                metric_name="row_count",
                metric_args=None,
                filter=None,
                aggregation=None,
                check_missing_and_valid_cfg=None,
                column_configurations_cfg=None,
                check=self,
            )
        )

        self.metrics["other_row_count"] = data_source_scan.resolve_metric(
            NumericQueryMetric(
                data_source_scan=data_source_scan,
                partition=self.other_partition,
                column=None,
                metric_name="row_count",
                metric_args=None,
                filter=None,
                aggregation=None,
                check_missing_and_valid_cfg=None,
                column_configurations_cfg=None,
                check=self,
            )
        )

    def get_cloud_diagnostics_dict(self) -> dict:
        row_count: int = self.metrics["row_count"].value
        other_row_count: int = self.metrics["other_row_count"].value

        return {
            # TODO Check with Soda Cloud what should be the value here
            "value": row_count - other_row_count,
            "rowCount": row_count,
            "otherRowCount": other_row_count,
        }

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        row_count: int = metrics["row_count"].value
        other_row_count: int = metrics["other_row_count"].value

        diagnostic_values = {
            "row_count": row_count,
            "other_row_count": other_row_count,
        }

        row_count_diff = row_count - other_row_count

        self.outcome = CheckOutcome.PASS if row_count_diff == 0 else CheckOutcome.FAIL
        self.check_value = row_count_diff
        self.diagnostic_values = diagnostic_values



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import os
from pathlib import Path

logger = logging.getLogger(__name__)


class FileSystem:
    def user_home_dir(self):
        return str(Path.home())

    def expand_user(self, path: str):
        return os.path.expanduser(path)

    def exists(self, path: str):
        return Path(self.expand_user(path)).exists()

    def is_dir(self, path: str):
        return Path(self.expand_user(path)).is_dir()

    def is_file(self, path: str):
        return Path(self.expand_user(path)).is_file()

    def file_read_as_str(self, path: str) -> str:
        with open(self.expand_user(path), encoding="utf-8") as f:
            return f.read()

    def file_write_from_str(self, path: str, file_content_str):
        path = self.expand_user(path)
        path_path: Path = Path(path)

        is_new = not path_path.exists()
        with open(path_path, "w+", encoding="utf-8") as f:
            f.write(file_content_str)
        if is_new:
            os.chmod(path, 0o666)

    def scan_dir(self, dir_path: str):
        return os.scandir(self.expand_user(dir_path))

    def dirname(self, path: str):
        return os.path.dirname(self.expand_user(path))

    def mkdirs(self, path: str, absolute: bool = False):
        Path(self.expand_user(path)).mkdir(parents=True, exist_ok=True)

    def file_write_from_str(self, path: str, file_content_str):
        expanded_path = os.path.expanduser(path)
        path_path: Path = Path(expanded_path)
        try:
            with open(path_path, "w+", encoding="utf-8") as f:
                f.write(file_content_str)
        except Exception as e:
            logger.debug(f"Couldn't write {str(path)}: {str(e)}")


class FileSystemSingleton:
    INSTANCE: FileSystem = FileSystem()


def file_system() -> FileSystem:
    return FileSystemSingleton.INSTANCE



from typing import Tuple

from soda.sampler.sample import Sample
from soda.sampler.sample_schema import SampleColumn, SampleSchema


class DbSample(Sample):
    def __init__(self, cursor, data_source):
        self.cursor = cursor
        self.data_source = data_source

    def get_rows(self) -> Tuple[Tuple]:
        return self.cursor.fetchall()

    def get_schema(self) -> SampleSchema:
        return self._convert_python_db_schema_to_sample_schema(self.cursor.description)

    def _convert_python_db_schema_to_sample_schema(self, description) -> SampleSchema:
        return SampleSchema(
            columns=[
                self._convert_python_db_column_to_sample_column(python_db_column) for python_db_column in description
            ]
        )

    def _convert_python_db_column_to_sample_column(self, python_db_column):
        type_code = python_db_column[1]
        type_name = self.data_source.get_type_name(type_code)
        return SampleColumn(name=python_db_column[0], type=type_name)



from datetime import datetime, timedelta, timezone
from typing import Dict, Optional

from soda.execution.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric import Metric
from soda.execution.numeric_query_metric import NumericQueryMetric

MAX_COLUMN_TIMESTAMP = "max_column_timestamp"


class FreshnessCheck(Check):
    def __init__(
        self,
        check_cfg: "FreshnessCheckCfg",
        data_source_scan: "DataSourceScan",
        partition: "Partition",
        column: "Column",
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
            name="Freshness",
        )
        self.freshness_values: Optional[dict] = None
        self.metrics[MAX_COLUMN_TIMESTAMP] = data_source_scan.resolve_metric(
            NumericQueryMetric(
                data_source_scan=self.data_source_scan,
                partition=self.partition,
                column=self.column,
                metric_name="max",
                metric_args=None,
                filter=None,
                aggregation=None,
                check_missing_and_valid_cfg=None,
                column_configurations_cfg=None,
                check=self,
            )
        )

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        from soda.sodacl.freshness_check_cfg import FreshnessCheckCfg

        check_cfg: FreshnessCheckCfg = self.check_cfg

        now_variable_name = check_cfg.variable_name
        now_variable_timestamp_text = self.data_source_scan.scan.get_variable(now_variable_name)
        if not now_variable_timestamp_text:
            self.outcome = CheckOutcome.FAIL
            self.logs.error(
                f"Could not parse variable {now_variable_name} as a timestamp: variable not found",
            )
            return

        max_column_timestamp: Optional[datetime] = metrics.get(MAX_COLUMN_TIMESTAMP).value
        now_variable_timestamp: Optional[datetime] = None

        is_max_column_timestamp_valid = isinstance(max_column_timestamp, datetime)
        if not is_max_column_timestamp_valid and max_column_timestamp is not None:
            self.logs.error(
                f"Could not evaluate freshness: max({self.check_cfg.column_name}) "
                f"is not a datetime: {type(max_column_timestamp).__name__}",
                location=self.check_cfg.location,
            )

        try:
            now_variable_timestamp = datetime.fromisoformat(now_variable_timestamp_text)
        except:
            self.logs.error(
                f"Could not parse variable {now_variable_name} as a timestamp: {now_variable_timestamp_text}",
                location=check_cfg.location,
            )

        is_now_variable_timestamp_valid = isinstance(now_variable_timestamp, datetime)

        freshness = None
        if not is_now_variable_timestamp_valid:
            self.logs.error(
                f"Could not evaluate freshness: variable {check_cfg.variable_name} is not a datetime: {now_variable_timestamp_text}",
                location=self.check_cfg.location,
            )

        max_column_timestamp_utc = (
            self._datetime_to_utc(max_column_timestamp) if isinstance(max_column_timestamp, datetime) else None
        )
        now_timestamp_utc = (
            self._datetime_to_utc(now_variable_timestamp) if isinstance(now_variable_timestamp, datetime) else None
        )

        if is_max_column_timestamp_valid and is_now_variable_timestamp_valid:
            freshness = now_timestamp_utc - max_column_timestamp_utc
            if check_cfg.fail_staleness_threshold is not None and freshness > check_cfg.fail_staleness_threshold:
                self.outcome = CheckOutcome.FAIL
            elif check_cfg.warn_staleness_threshold is not None and freshness > check_cfg.warn_staleness_threshold:
                self.outcome = CheckOutcome.WARN
            else:
                self.outcome = CheckOutcome.PASS
        elif max_column_timestamp is None:
            self.outcome = CheckOutcome.FAIL

        self.freshness_values = {
            "max_column_timestamp": str(max_column_timestamp) if max_column_timestamp else None,
            "max_column_timestamp_utc": str(max_column_timestamp_utc) if max_column_timestamp_utc else None,
            "now_variable_name": now_variable_name,
            "now_timestamp": now_variable_timestamp_text,
            "now_timestamp_utc": str(now_timestamp_utc) if now_timestamp_utc else None,
            "freshness": freshness,
        }

    def get_cloud_diagnostics_dict(self):
        freshness_dict = {}

        freshness = 0
        if self.freshness_values["freshness"] and isinstance(self.freshness_values["freshness"], timedelta):
            freshness = self.freshness_values["freshness"].microseconds
        freshness_dict = {
            "value": freshness,  # microseconds difference
            "maxColumnTimestamp": self.freshness_values["max_column_timestamp"],
            "maxColumnTimestampUtc": self.freshness_values["max_column_timestamp_utc"],
            "nowVariableName": self.freshness_values["now_variable_name"],
            "nowTimestamp": self.freshness_values["now_timestamp"],
            "nowTimestampUtc": self.freshness_values["now_timestamp_utc"],
            "freshness": self.freshness_values["freshness"],
        }

        return freshness_dict

    def get_log_diagnostic_dict(self) -> dict:
        return self.freshness_values

    @staticmethod
    def _datetime_to_utc(input: datetime) -> datetime:
        if input.tzinfo is None:
            return input.replace(tzinfo=timezone.utc)
        return input.astimezone(timezone.utc)



from typing import Dict, List, Optional

from soda.execution.metric import Metric

METRIC_NAME_MISSING_PERCENT = "missing_percent"
METRIC_NAME_INVALID_PERCENT = "invalid_percent"
DERIVED_METRIC_NAMES = [
    METRIC_NAME_MISSING_PERCENT,
    METRIC_NAME_INVALID_PERCENT,
]


class DerivedMetric(Metric):
    def __init__(
        self,
        data_source_scan: "DataSourceScan",
        partition: Optional["Partition"],
        column: Optional["Column"],
        metric_name: str,
        metric_args: Optional[List[object]],
        filter: str,
        check_missing_and_valid_cfg: "MissingAndValidCfg",
        column_configurations_cfg: "ColumnConfigurationsCfg",
        check: "Check",
    ):
        from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg

        merged_missing_and_valid_cfg = MissingAndValidCfg.merge(check_missing_and_valid_cfg, column_configurations_cfg)
        other_metric_args = metric_args[1:] if isinstance(metric_args, list) and len(metric_args) > 1 else None
        super().__init__(
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
            name=metric_name,
            check=check,
            identity_parts=[
                other_metric_args,
                filter,
                merged_missing_and_valid_cfg.get_identity_parts() if merged_missing_and_valid_cfg else None,
            ],
        )

        from soda.execution.derived_formula import DerivedFormula

        # Metric identity fields
        # The properties that logically define the metric
        # (this logical identity is actually implemented with Python equality see __eq__,
        # and hence different from Python identity)

        self.metric_name: Optional[str] = metric_name
        self.metric_args: Optional[List[object]] = metric_args
        self.filter: str = filter
        self.missing_and_valid_cfg: MissingAndValidCfg = merged_missing_and_valid_cfg
        self.derived_formula: DerivedFormula = self.create_percentage_formula()

    def create_percentage_formula(self):
        from soda.execution.derived_formula import DerivedFormula

        if METRIC_NAME_MISSING_PERCENT == self.name:

            def derive_missing_percentage(values: Dict[str, int]):
                missing_percentage = values["missing_count"] * 100 / values["row_count"]
                return float(round(missing_percentage, 2))

            return DerivedFormula(
                function=derive_missing_percentage,
                metric_dependencies={
                    "row_count": self.build_dependency_metric("row_count"),
                    "missing_count": self.build_dependency_metric("missing_count"),
                },
            )

        if METRIC_NAME_INVALID_PERCENT == self.name:

            def derive_invalid_percentage(values: Dict[str, int]):
                invalid_percentage = values["invalid_count"] * 100 / values["row_count"]
                return float(round(invalid_percentage, 2))

            return DerivedFormula(
                function=derive_invalid_percentage,
                metric_dependencies={
                    "row_count": self.build_dependency_metric("row_count"),
                    "invalid_count": self.build_dependency_metric("invalid_count"),
                },
            )

    def build_dependency_metric(self, metric_name: str) -> "Metric":
        from soda.execution.numeric_query_metric import NumericQueryMetric

        return self.data_source_scan.resolve_metric(
            NumericQueryMetric(
                data_source_scan=self.data_source_scan,
                partition=self.partition,
                column=self.column if metric_name != "row_count" else None,
                metric_name=metric_name,
                metric_args=self.metric_args,
                filter=self.filter,
                aggregation=None,
                check_missing_and_valid_cfg=self.missing_and_valid_cfg if metric_name != "row_count" else None,
                column_configurations_cfg=None,
                check=next(iter(self.checks)),
            )
        )

    def compute_derived_metric_values(self):
        """
        Resolves derived metrics recursively.
        Query metric values are all found in parameter metric_values.
        As derived metrics are resolved, the new derived metric values are added to metric_values.
        """

        for dependency_metric in self.derived_formula.metric_dependencies.values():
            if isinstance(dependency_metric, DerivedMetric):
                dependency_metric.compute_derived_metric_values()

        formula_values: Dict[str, object] = {}
        for (
            metric_dependency_name,
            dependency_metric,
        ) in self.derived_formula.metric_dependencies.items():
            formula_values[metric_dependency_name] = dependency_metric.value

        self.value = self.derived_formula.function(formula_values)
        self.formula_values = formula_values
        failed_rows_sample_refs = self.derived_formula.collect_failed_rows_sample_refs()
        if len(failed_rows_sample_refs) == 1:
            self.failed_rows_sample_ref = failed_rows_sample_refs[0]



from __future__ import annotations


class DiscoverTablesResultColumn:
    def __init__(self, column_name: str, column_type: str):
        self.column_name: str = column_name
        self.column_type: str = column_type

    def get_cloud_dict(self) -> dict:
        cloud_dict = {
            "columnName": self.column_name,
            "sourceDataType": self.column_type,
        }
        return cloud_dict



from __future__ import annotations

import logging
import os
import re
from builtins import isinstance
from datetime import timedelta
from numbers import Number
from textwrap import dedent

from antlr4 import CommonTokenStream, InputStream
from antlr4.error.ErrorListener import ErrorListener
from soda.common.logs import Logs
from soda.common.parser import Parser
from soda.common.yaml_helper import to_yaml_str
from soda.sodacl.antlr.SodaCLAntlrLexer import SodaCLAntlrLexer
from soda.sodacl.antlr.SodaCLAntlrParser import SodaCLAntlrParser
from soda.sodacl.automated_monitoring_cfg import AutomatedMonitoringCfg
from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg
from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.discover_tables_cfg import DiscoverTablesCfg
from soda.sodacl.distribution_check_cfg import DistributionCheckCfg
from soda.sodacl.for_each_column_cfg import ForEachColumnCfg
from soda.sodacl.for_each_table_cfg import ForEachTableCfg
from soda.sodacl.freshness_check_cfg import FreshnessCheckCfg
from soda.sodacl.missing_and_valid_cfg import CFG_MISSING_VALID_ALL, MissingAndValidCfg
from soda.sodacl.name_filter import NameFilter
from soda.sodacl.profile_columns_cfg import ProfileColumnsCfg
from soda.sodacl.reference_check_cfg import ReferenceCheckCfg
from soda.sodacl.row_count_comparison_check_cfg import RowCountComparisonCheckCfg
from soda.sodacl.schema_check_cfg import SchemaCheckCfg, SchemaValidations
from soda.sodacl.sodacl_cfg import SodaCLCfg
from soda.sodacl.table_cfg import TableCfg
from soda.sodacl.threshold_cfg import ThresholdCfg

logger = logging.getLogger(__name__)


WARN = "warn"
FAIL = "fail"
NAME = "name"
IDENTITY = "identity"
FAIL_CONDITION = "fail condition"
FAIL_QUERY = "fail query"
WHEN_REQUIRED_COLUMN_MISSING = "when required column missing"
WHEN_WRONG_COLUMN_TYPE = "when wrong column type"
WHEN_WRONG_COLUMN_INDEX = "when wrong column index"
WHEN_FORBIDDEN_COLUMN_PRESENT = "when forbidden column present"
WHEN_SCHEMA_CHANGES = "when schema changes"
ALL_SCHEMA_VALIDATIONS = [
    WHEN_REQUIRED_COLUMN_MISSING,
    WHEN_WRONG_COLUMN_TYPE,
    WHEN_WRONG_COLUMN_INDEX,
    WHEN_FORBIDDEN_COLUMN_PRESENT,
    WHEN_SCHEMA_CHANGES,
]


class SodaCLParser(Parser):
    def __init__(
        self,
        sodacl_cfg: SodaCLCfg,
        logs: Logs,
        file_path: str,
        data_source_name: str,
    ):
        super().__init__(file_path=file_path, logs=logs)
        from soda.sodacl.sodacl_cfg import SodaCLCfg

        self.sodacl_cfg: SodaCLCfg = sodacl_cfg
        self.data_source_name = data_source_name

    def parse_sodacl_yaml_str(self, sodacl_yaml_str: str):
        sodacl_dict = self._parse_yaml_str(sodacl_yaml_str)
        if sodacl_dict is not None:
            self.__parse_headers(sodacl_dict)

    def __parse_headers(self, headers_dict: dict) -> None:
        if not headers_dict:
            return

        for header_str, header_content in headers_dict.items():
            self._push_path_element(header_str, header_content)
            try:
                if "automated monitoring" == header_str:
                    self.__parse_automated_monitoring_section(header_str, header_content)
                elif header_str.startswith("profile columns"):
                    self.__parse_profile_columns_section(header_str, header_content)
                elif header_str.startswith("discover tables"):
                    self.__parse_discover_tables_section(header_str, header_content)
                elif "checks" == header_str:
                    self.__parse_data_source_checks_section(header_str, header_content)
                elif "variables" == header_str:
                    self.__parse_header_section(header_str, header_content)
                else:
                    antlr_parser = self.antlr_parse_section_header(header_str)
                    if antlr_parser.is_ok():
                        antlr_section_header = antlr_parser.result

                        if antlr_section_header.table_checks_header():
                            self.__parse_table_checks_section(
                                antlr_section_header.table_checks_header(),
                                header_str,
                                header_content,
                            )
                        elif antlr_section_header.column_configurations_header():
                            self.__parse_column_configurations_section(
                                antlr_section_header.column_configurations_header(),
                                header_str,
                                header_content,
                            )
                        elif antlr_section_header.table_filter_header():
                            self.__parse_table_filter_section(
                                antlr_section_header.table_filter_header(), header_str, header_content
                            )
                        elif antlr_section_header.checks_for_each_table_header():
                            self.__parse_antlr_checks_for_each_table_section(
                                antlr_section_header.checks_for_each_table_header(),
                                header_str,
                                header_content,
                            )
                        elif antlr_section_header.checks_for_each_column_header():
                            self.__parse_antlr_checks_for_each_column_section(
                                antlr_section_header.checks_for_each_column_header(),
                                header_str,
                                header_content,
                            )
                        else:
                            self.logs.error(
                                f'Skipping unknown section header "{header_str}"',
                                location=self.location,
                            )
                    else:
                        self.logs.error(
                            f'Invalid section header "{header_str}"',
                            location=self.location,
                        )

            finally:
                self._pop_path_element()

    def __parse_table_checks_section(self, antlr_table_checks_header, header_str, header_content):
        if isinstance(header_content, list):
            table_name = self.__antlr_parse_identifier_name_from_header(antlr_table_checks_header)
            if table_name is None:
                self.logs.error(
                    f'No table name in section header "{header_str}"',
                    location=self.location,
                )

            data_source_scan_cfg = self.sodacl_cfg._get_or_create_data_source_scan_cfgs(self.data_source_name)
            table_cfg: TableCfg = data_source_scan_cfg.get_or_create_table_cfg(table_name)

            partition_name = self.__antlr_parse_partition_from_header(antlr_table_checks_header)
            partition_cfg = table_cfg.find_partition(self.location.file_path, partition_name)
            if partition_cfg is None:
                self.logs.error(f"Filter not declared: {partition_name}", location=self.location)
                return None

            partition_cfg.locations.append(self.location)

            for check_index, check_list_element in enumerate(header_content):
                self._push_path_element(check_index, check_list_element)

                check_str: str = None
                check_configurations = None

                if isinstance(check_list_element, str):
                    check_str = check_list_element
                elif isinstance(check_list_element, dict):
                    check_str = next(iter(check_list_element))
                    check_configurations = check_list_element[check_str]
                else:
                    self.logs.error(
                        f"Skipping unsupported check definition: {to_yaml_str(check_list_element)}",
                        location=self.location,
                    )

                if check_str is not None:
                    check_cfg = self.__parse_table_check_str(header_str, check_str, check_configurations)

                    if check_cfg:
                        column_name = check_cfg.get_column_name()
                        if column_name:
                            column_checks = partition_cfg.get_or_create_column_checks(column_name)
                            column_checks.add_check_cfg(check_cfg)
                        else:
                            partition_cfg.add_check_cfg(check_cfg)

                self._pop_path_element()
        else:
            self.logs.error(
                f'Skipping section "{header_str}" because content is not a list',
                location=self.location,
            )

    def __parse_data_source_checks_section(self, header_str, header_content):
        if isinstance(header_content, list):
            data_source_scan_cfg = self.sodacl_cfg._get_or_create_data_source_scan_cfgs(self.data_source_name)

            for check_index, check_list_element in enumerate(header_content):
                self._push_path_element(check_index, check_list_element)

                check_str: str = None
                check_configurations = None

                if isinstance(check_list_element, str):
                    check_str = check_list_element
                elif isinstance(check_list_element, dict):
                    check_str = next(iter(check_list_element))
                    check_configurations = check_list_element[check_str]
                else:
                    self.logs.error(
                        f"Skipping unsupported check definition: {to_yaml_str(check_list_element)}",
                        location=self.location,
                    )

                if check_str is not None:
                    check_cfg = self.__parse_data_source_check_str(header_str, check_str, check_configurations)

                    if check_cfg:
                        data_source_scan_cfg.add_check_cfg(check_cfg)
                    else:
                        self.logs.error(
                            f"Could not detect check type of '{check_str}'",
                            location=self.location,
                        )

                self._pop_path_element()
        else:
            self.logs.error(
                f'Skipping section "{header_str}" because content is not a list',
                location=self.location,
            )

    def __parse_table_check_str(self, header_str: str, check_str: str, check_configurations) -> CheckCfg:
        if check_str == "schema":
            return self.__parse_schema_check(header_str, check_str, check_configurations)

        elif check_str == "failed rows":
            return self.parse_user_defined_failed_rows_check_cfg(check_configurations, check_str, header_str)

        else:
            antlr_parser = self.antlr_parse_check(check_str)
            if antlr_parser.is_ok():
                antlr_check = antlr_parser.result
                if antlr_check.metric_check():
                    return self.__parse_metric_check(
                        antlr_check.metric_check(),
                        header_str,
                        check_str,
                        check_configurations,
                    )

                elif antlr_check.row_count_comparison_check():
                    return self.__parse_row_count_comparison_check(
                        antlr_check.row_count_comparison_check(),
                        header_str,
                        check_str,
                        check_configurations,
                    )

                elif antlr_check.reference_check():
                    return self.__parse_reference_check(
                        antlr_check.reference_check(),
                        header_str,
                        check_str,
                        check_configurations,
                    )

                elif antlr_check.freshness_check():
                    return self.__parse_freshness_check(
                        antlr_check.freshness_check(),
                        header_str,
                        check_str,
                        check_configurations,
                    )
            else:
                self.logs.error(f'Invalid check "{check_str}": {antlr_parser.error_message}')

    def __parse_data_source_check_str(
        self,
        header_str: str,
        check_str: str,
        check_configurations: dict | None,
    ) -> CheckCfg:
        if check_str == "failed rows":
            return self.parse_failed_rows_data_source_query_check(header_str, check_str, check_configurations)

        else:
            antlr_parser = self.antlr_parse_check(check_str)
            if antlr_parser.is_ok():
                antlr_check = antlr_parser.result
                if antlr_check.metric_check():
                    return self.__parse_metric_check(
                        antlr_check.metric_check(),
                        header_str,
                        check_str,
                        check_configurations,
                    )
                else:
                    self.logs.error(f'Check type "{check_str}" must be in "checks for TABLENAME:" section.')
            else:
                self.logs.error(f'Invalid check "{check_str}": {antlr_parser.error_message}')

    def __parse_header_section(self, header_str, header_content):
        if isinstance(header_content, dict):
            variables = self.sodacl_cfg.scan._variables
            for variable_name in header_content:
                variable_value = header_content[variable_name]
                if "${" in variable_value:
                    variable_value = self._resolve_jinja(variable_value, variables)
                variables[variable_name] = variable_value
        else:
            self.logs.error(f"Variables content must be a dict.  Was {type(header_content).__name__}")

    def parse_user_defined_failed_rows_check_cfg(self, check_configurations, check_str, header_str):
        if isinstance(check_configurations, dict):
            from soda.sodacl.user_defined_failed_rows_check_cfg import (
                UserDefinedFailedRowsCheckCfg,
            )
            from soda.sodacl.user_defined_failed_rows_expression_check_cfg import (
                UserDefinedFailedRowsExpressionCheckCfg,
            )

            self._push_path_element(check_str, check_configurations)
            try:
                name = self._get_required(NAME, str)
                for invalid_configuration_key in [
                    key for key in check_configurations if key not in [NAME, WARN, FAIL, FAIL_CONDITION, FAIL_QUERY]
                ]:
                    self.logs.error(
                        f'Invalid user defined failed rows check configuration key "{invalid_configuration_key}"',
                        location=self.location,
                    )
                fail_condition_sql_expr = self._get_optional(FAIL_CONDITION, str)
                if fail_condition_sql_expr:
                    return UserDefinedFailedRowsExpressionCheckCfg(
                        source_header=header_str,
                        source_line=check_str,
                        source_configurations=check_configurations,
                        location=self.location,
                        name=name,
                        fail_condition_sql_expr=fail_condition_sql_expr,
                    )
                else:
                    fail_query = self._get_optional(FAIL_QUERY, str)
                    if fail_query:
                        return UserDefinedFailedRowsCheckCfg(
                            source_header=header_str,
                            source_line=check_str,
                            source_configurations=check_configurations,
                            location=self.location,
                            name=name,
                            query=fail_query,
                        )
                    else:
                        self.logs.error(
                            'In a "failed rows" check, either "fail condition" or "fail query" are required as nested configurations.',
                            location=self.location,
                        )

            finally:
                self._pop_path_element()
        else:
            self.logs.error(f'Check "{check_str}" expects a nested object/dict, but was {check_configurations}')

    def __parse_failed_rows_table_expression_check(
        self,
        header_str: str,
        check_str: str,
        check_configurations: dict | None,
    ) -> CheckCfg:
        if isinstance(check_configurations, dict):
            from soda.sodacl.user_defined_failed_rows_expression_check_cfg import (
                UserDefinedFailedRowsExpressionCheckCfg,
            )

            self._push_path_element(check_str, check_configurations)
            try:
                expression = self._get_required("failed rows expression", str)
                name = self._get_optional(NAME, str)
                return UserDefinedFailedRowsExpressionCheckCfg(
                    source_header=header_str,
                    source_line=check_str,
                    source_configurations=check_configurations,
                    location=self.location,
                    name=name,
                    fail_condition_sql_expr=expression,
                )
            finally:
                self._pop_path_element()
        else:
            self.logs.error(f'Check "{check_str}" expects a nested object/dict, but was {check_configurations}')

    def parse_failed_rows_data_source_query_check(
        self,
        header_str: str,
        check_str: str,
        check_configurations: dict | None,
    ) -> CheckCfg:
        from soda.sodacl.user_defined_failed_rows_check_cfg import (
            UserDefinedFailedRowsCheckCfg,
        )

        if isinstance(check_configurations, dict):
            self._push_path_element(check_str, check_configurations)
            try:
                name = self._get_optional(NAME, str)
                query = self._get_required(FAIL_QUERY, str)
                return UserDefinedFailedRowsCheckCfg(
                    source_header=header_str,
                    source_line=check_str,
                    source_configurations=check_configurations,
                    location=self.location,
                    name=name,
                    query=query,
                )
            finally:
                self._pop_path_element()
        else:
            self.logs.error(
                f'"failed rows" check must have configurations',
                location=self.location,
            )

    def __parse_metric_check(
        self,
        antlr_metric_check,
        header_str: str,
        check_str: str,
        check_configurations: dict | None,
    ) -> CheckCfg:
        from soda.sodacl.metric_check_cfg import MetricCheckCfg

        antlr_metric = antlr_metric_check.metric()
        metric_name = antlr_metric.metric_name().getText()
        metric_args = None
        if antlr_metric.metric_args():
            metric_args = [
                self.__parse_metric_arg(metric_arg)
                for metric_arg in antlr_metric.metric_args().getChildren()
                if isinstance(metric_arg, SodaCLAntlrParser.Metric_argContext)
            ]

        antlr_threshold = antlr_metric_check.threshold()
        fail_threshold_cfg = None
        warn_threshold_cfg = None
        if antlr_threshold:
            pass_threshold_cfg = self.__antlr_parse_threshold_condition(antlr_threshold)
            fail_threshold_cfg = pass_threshold_cfg.get_inverse()
        elif isinstance(check_configurations, dict):
            self._push_path_element(check_str, check_configurations)
            fail_threshold_condition_str = self._get_optional(FAIL, str)
            fail_threshold_cfg = self.__parse_configuration_threshold_condition(fail_threshold_condition_str)
            warn_threshold_condition_str = self._get_optional(WARN, str)
            warn_threshold_cfg = self.__parse_configuration_threshold_condition(warn_threshold_condition_str)
            self._pop_path_element()

        # Parse the nested check configuration details
        name = None
        filter = None
        missing_and_valid_cfg = None
        condition = None
        metric_expression = None
        metric_query = None

        if isinstance(check_configurations, dict):
            for configuration_key in check_configurations:
                configuration_value = check_configurations[configuration_key]
                self._push_path_element(check_str, check_configurations)
                name = self._get_optional(NAME, str)
                self._push_path_element(configuration_key, configuration_value)
                if "filter" == configuration_key:
                    filter = configuration_value.strip()
                elif "condition" == configuration_key:
                    condition = configuration_value.strip()
                elif configuration_key.endswith("expression"):
                    metric_expression = configuration_value.strip()
                    configuration_metric_name = (
                        configuration_key[: -len(" expression")]
                        if len(configuration_key) > len(" expression")
                        else None
                    )
                    if configuration_metric_name != metric_name:
                        self.logs.error(
                            f'In configuration "{configuration_key}" the metric name must match exactly the metric name in the check "{metric_name}"',
                            location=self.location,
                        )
                elif configuration_key.endswith("query"):
                    metric_query = dedent(configuration_value).strip()
                    configuration_metric_name = (
                        configuration_key[: -len(" query")] if len(configuration_key) > len(" query") else None
                    )
                    if configuration_metric_name != metric_name:
                        self.logs.error(
                            f'In configuration "{configuration_key}" the metric name must match exactly the metric name in the check "{metric_name}"',
                            location=self.location,
                        )
                elif configuration_key in CFG_MISSING_VALID_ALL:
                    if missing_and_valid_cfg is None:
                        missing_and_valid_cfg = MissingAndValidCfg()
                    self.__parse_missing_and_valid(
                        configuration_key,
                        configuration_value,
                        missing_and_valid_cfg,
                    )
                elif configuration_key not in [NAME, IDENTITY, WARN, FAIL]:
                    if metric_name != "distribution_difference":
                        self.logs.error(
                            f"Skipping unsupported check configuration: {configuration_key}",
                            doc="metric_check_configurations",
                            location=self.location,
                        )

                self._pop_path_element()
                self._pop_path_element()

        elif check_configurations is not None:
            self.logs.error(
                f"Check configuration details should be object/dict, but was {type(check_configurations).__name__}: \n{to_yaml_str(check_configurations)}",
                location=self.location,
            )

        metric_check_cfg_class = MetricCheckCfg

        change_over_time_cfg = None
        if antlr_metric_check.change_over_time():
            from soda.sodacl.change_over_time_metric_check_cfg import (
                ChangeOverTimeMetricCheckCfg,
            )

            metric_check_cfg_class = ChangeOverTimeMetricCheckCfg
            change_over_time_cfg = ChangeOverTimeCfg()
            antlr_change_over_time = antlr_metric_check.change_over_time()
            change_over_time_config = antlr_change_over_time.change_over_time_config()
            if change_over_time_config:
                if change_over_time_config.LAST():
                    change_over_time_cfg.last_measurements = int(change_over_time_config.integer().getText())
                    change_over_time_cfg.last_aggregation = change_over_time_config.change_aggregation().getText()
                elif change_over_time_config.SAME_DAY_LAST_WEEK():
                    change_over_time_cfg.same_day_last_week = True
                elif change_over_time_config.SAME_DAY_LAST_MONTH():
                    change_over_time_cfg.same_day_last_month = True
            else:
                change_over_time_cfg.last_measurements = 1
                change_over_time_cfg.last_aggregation = "min"

        if antlr_metric_check.anomaly_score():
            from soda.sodacl.anomaly_metric_check_cfg import AnomalyMetricCheckCfg

            metric_check_cfg_class = AnomalyMetricCheckCfg

            if not antlr_metric_check.default_anomaly_threshold():
                self.logs.error(
                    "Non default threshold for anomaly detection is currently not supported. "
                    "Make sure that your check definition adheres to the following syntax: 'anomaly score for {metric} < default'.",
                    location=self.location,
                )
                return metric_check_cfg_class(
                    source_header=header_str,
                    source_line=check_str,
                    source_configurations=check_configurations,
                    location=self.location,
                    name=name,
                    metric_name=metric_name,
                    metric_args=metric_args,
                    missing_and_valid_cfg=missing_and_valid_cfg,
                    filter=filter,
                    condition=condition,
                    metric_expression=metric_expression,
                    metric_query=metric_query,
                    change_over_time_cfg=change_over_time_cfg,
                    fail_threshold_cfg=None,
                    warn_threshold_cfg=None,
                )
            if antlr_metric_check.default_anomaly_threshold():
                warn_threshold_cfg = ThresholdCfg(gt=0.9)

            # NOTE: Right now we are not using this parsing as we require '< default' syntax but when that changes
            # this section will come in handy to parse user-provided thresholds for anomaly detection.
            elif antlr_threshold:
                warn_threshold_cfg = fail_threshold_cfg
                fail_threshold_cfg = None

            if (fail_threshold_cfg and not fail_threshold_cfg.is_valid_anomaly_threshold()) or (
                warn_threshold_cfg and not warn_threshold_cfg.is_valid_anomaly_threshold()
            ):
                if antlr_threshold:
                    self.logs.error(
                        'Invalid anomaly threshold.  Only "< default" or "< {threshold-value}" '
                        "are allowed where threshold-value must be between 0 (ok) and 1 (anomaly)",
                        location=self.location,
                    )
                else:
                    self.logs.error(
                        'Invalid anomaly threshold.  Only "when > {threshold-value}" '
                        "is allowed where threshold-value must be between 0 (ok) and 1 (anomaly)",
                        location=self.location,
                    )

        elif antlr_metric_check.default_anomaly_threshold():
            self.logs.error(
                'Threshold "< default" only allowed for anomaly checks that start with: "anomaly score '
                "{metric} < {threshold}",
                location=self.location,
            )
        elif metric_name == "distribution_difference":
            column_name: str = metric_args[0]
            distribution_name: str = metric_args[1]
            reference_file_path: str = os.path.join(
                os.path.dirname(self.location.file_path), check_configurations.get("distribution reference file")
            )
            return DistributionCheckCfg(
                source_header=header_str,
                source_line=check_str,
                source_configurations=check_configurations,
                location=self.location,
                name=name,
                column_name=column_name,
                distribution_name=distribution_name,
                filter=None,
                reference_file_path=reference_file_path,
                fail_threshold_cfg=fail_threshold_cfg,
                warn_threshold_cfg=warn_threshold_cfg,
            )

        if fail_threshold_cfg is None and warn_threshold_cfg is None:
            self.logs.error(f'No threshold specified for check "{check_str}"', location=self.location)

        return metric_check_cfg_class(
            source_header=header_str,
            source_line=check_str,
            source_configurations=check_configurations,
            location=self.location,
            name=name,
            metric_name=metric_name,
            metric_args=metric_args,
            missing_and_valid_cfg=missing_and_valid_cfg,
            filter=filter,
            condition=condition,
            metric_expression=metric_expression,
            metric_query=metric_query,
            change_over_time_cfg=change_over_time_cfg,
            fail_threshold_cfg=fail_threshold_cfg,
            warn_threshold_cfg=warn_threshold_cfg,
        )

    def __parse_configuration_threshold_condition(self, value) -> ThresholdCfg | None:
        if isinstance(value, str):
            if not value.startswith("when "):
                self.logs.error(
                    'Value for fail must be "when {condition}"',
                    location=self.location,
                )
            else:
                threshold_str = value[len("when ") :]
                antlr_parser = self.antlr_parse_threshold(threshold_str)
                if antlr_parser.is_ok():
                    antlr_threshold = antlr_parser.result
                    return self.__antlr_parse_threshold_condition(antlr_threshold)
                else:
                    self.logs.error(f'Invalid threshold "{threshold_str}": {antlr_parser.error_message}')

    def __parse_metric_arg(self, antlr_metric_arg):
        antlr_signed_number = antlr_metric_arg.signed_number()
        if antlr_signed_number:
            return self.__antlr_parse_signed_number(antlr_signed_number)
        return antlr_metric_arg.identifier().getText()

    def __parse_column_configurations_section(self, antlr_configuration_header, header_str, header_content):
        table_name = self.__antlr_parse_identifier_name_from_header(antlr_configuration_header)
        if table_name is None:
            self.logs.error(
                f'Table not specified in "{header_str}"',
                location=self.location,
            )
            return

        if isinstance(header_content, dict):
            data_source_scan_cfg = self.sodacl_cfg._get_or_create_data_source_scan_cfgs(self.data_source_name)
            table_cfg: TableCfg = data_source_scan_cfg.get_or_create_table_cfg(table_name)
            table_cfg.column_configuration_locations.append(self.location)

            for (
                configuration_key,
                configuration_value,
            ) in header_content.items():
                parts = configuration_key.split(" for ")
                if len(parts) != 2:
                    self.logs.error(
                        f"Column configuration key {configuration_key} not in appropriate format {{configuration_type}} for {{column}}",
                        location=self.location,
                    )
                else:
                    configuration_type = parts[0].strip()
                    column_name = parts[1].strip()
                    column_configurations_cfg = table_cfg.get_or_create_column_configurations(column_name)
                    if configuration_type in CFG_MISSING_VALID_ALL:
                        self.__parse_missing_and_valid(
                            configuration_type,
                            configuration_value,
                            column_configurations_cfg,
                        )
                    else:
                        self.logs.error(f'Invalid configuration "{configuration_key}"', location=self.location)
        elif isinstance(header_content, list) and len(header_content) > 0:
            self.logs.error(
                f'Contents of column configurations "{header_str}" is a list, but should be plain configurations.  Remove the "-" before the nested configurations',
                location=self.location,
            )
        else:
            self.logs.error(
                f'Skipping section "{header_str}" because content is not a object/dict',
                location=self.location,
            )

    def __parse_schema_check(self, header_str, check_str, check_configurations) -> SchemaCheckCfg | None:
        if isinstance(check_configurations, dict):
            self._push_path_element(check_str, check_configurations)
            for configuration_key in check_configurations:
                if configuration_key not in [NAME, WARN, FAIL]:
                    self.logs.error(
                        f'Invalid schema check configuration key "{configuration_key}"', location=self.location
                    )
            name = self._get_optional(NAME, str)
            schema_check_cfg = SchemaCheckCfg(
                source_header=header_str,
                source_line=check_str,
                source_configurations=check_configurations,
                location=self.location,
                name=name,
                warn_validations=self.__parse_schema_validations(WARN),
                fail_validations=self.__parse_schema_validations(FAIL),
            )
            self._pop_path_element()
            return schema_check_cfg
        else:
            self.logs.error(f'Check "{check_str}" expects a nested object/dict, but was {check_configurations}')

    def __parse_schema_validations(self, outcome_text: str):
        validations_dict = self._get_optional(outcome_text, dict)
        if validations_dict:
            self._push_path_element(outcome_text, validations_dict)

            is_column_addition_forbidden = False
            is_column_deletion_forbidden = False
            is_column_type_change_forbidden = False
            is_column_index_change_forbidden = False
            changes_not_allowed = validations_dict.get("when schema changes")
            if changes_not_allowed == "any":
                is_column_addition_forbidden = True
                is_column_deletion_forbidden = True
                is_column_type_change_forbidden = True
                is_column_index_change_forbidden = True
            elif isinstance(changes_not_allowed, list):
                for change_not_allowed in changes_not_allowed:
                    if change_not_allowed == "column add":
                        is_column_addition_forbidden = True
                    elif change_not_allowed == "column delete":
                        is_column_deletion_forbidden = True
                    elif change_not_allowed == "column type change":
                        is_column_type_change_forbidden = True
                    elif change_not_allowed == "column index change":
                        is_column_index_change_forbidden = True
                    else:
                        self.logs.error(f'"when schema changes" has invalid value {change_not_allowed}')
            elif changes_not_allowed is not None:
                self.logs.error(
                    f'Value for "when schema changes" must be either "any" or a list of these optioonal strings: {"column add", "column delete", "column type change", "column index change"}. Was {changes_not_allowed}'
                )

            schema_validations = SchemaValidations(
                required_column_names=self.__parse_schema_validation(WHEN_REQUIRED_COLUMN_MISSING),
                required_column_types=self.__parse_schema_validation(WHEN_WRONG_COLUMN_TYPE),
                required_column_indexes=self.__parse_schema_validation(WHEN_WRONG_COLUMN_INDEX),
                forbidden_column_names=self.__parse_schema_validation(WHEN_FORBIDDEN_COLUMN_PRESENT),
                is_column_addition_forbidden=is_column_addition_forbidden,
                is_column_deletion_forbidden=is_column_deletion_forbidden,
                is_column_type_change_forbidden=is_column_type_change_forbidden,
                is_column_index_change_forbidden=is_column_index_change_forbidden,
            )
            for invalid_schema_validation in [
                v
                for v in validations_dict
                if v
                not in [
                    WHEN_REQUIRED_COLUMN_MISSING,
                    WHEN_WRONG_COLUMN_TYPE,
                    WHEN_WRONG_COLUMN_INDEX,
                    WHEN_FORBIDDEN_COLUMN_PRESENT,
                    WHEN_SCHEMA_CHANGES,
                ]
            ]:
                hint = f"Available schema validations: {ALL_SCHEMA_VALIDATIONS}"
                if invalid_schema_validation == "when schema change":
                    hint = f'Did you mean "when schema changes" (plural)? {hint}'
                elif invalid_schema_validation == "when required columns missing":
                    hint = f'Did you mean "when required column missing"? (column in singular form) {hint}'
                elif invalid_schema_validation == "when forbidden columns present":
                    hint = f'Did you mean "when forbidden column present"? (column in singular form) {hint}'
                self.logs.error(
                    f'Invalid schema validation "{invalid_schema_validation}": {hint}',
                    location=self.location,
                )
            self._pop_path_element()
            return schema_validations

    def __parse_schema_validation(self, validation_type):
        value_type = (
            list
            if validation_type
            in [
                "when required column missing",
                "when forbidden column present",
            ]
            else dict
        )
        configuration_value = self._get_optional(validation_type, value_type)

        if configuration_value:
            if validation_type in [
                "when required column missing",
                "when forbidden column present",
            ]:
                are_values_valid = all(isinstance(c, str) for c in configuration_value)
            elif validation_type == "when wrong column type":
                are_values_valid = all(
                    isinstance(k, str) and isinstance(v, str) for k, v in configuration_value.items()
                )
            else:
                are_values_valid = all(
                    isinstance(k, str) and isinstance(v, int) for k, v in configuration_value.items()
                )
            if are_values_valid:
                return configuration_value
            else:
                self._push_path_element(validation_type, None)
                expected_configuration_type = (
                    "list of strings"
                    if validation_type
                    in [
                        "when required column missing",
                        "when forbidden column present",
                    ]
                    else "dict with strings for keys and values"
                    if validation_type == "when wrong column type"
                    else "dict with strings for keys and ints for values"
                )
                self.logs.error(
                    f'"{validation_type}" must contain {expected_configuration_type}',
                    location=self.location,
                )
                self._pop_path_element()

    def __parse_row_count_comparison_check(
        self,
        antlr_row_count_comparison_check,
        header_str: str,
        check_str: str,
        check_configurations: dict | None,
    ) -> CheckCfg:
        other_table_name = antlr_row_count_comparison_check.identifier(0).getText()
        antlr_identifier2 = antlr_row_count_comparison_check.identifier(1)

        other_partition_name = None
        antlr_partition_name = antlr_row_count_comparison_check.partition_name()
        if antlr_partition_name:
            other_partition_name = self.__antlr_parse_identifier(antlr_partition_name)

        other_data_source_name = antlr_identifier2.getText() if antlr_identifier2 else None

        name = None
        if isinstance(check_configurations, dict):
            self._push_path_element(check_str, check_configurations)
            name = self._get_optional(NAME, str)
            for configuration_key in check_configurations:
                if configuration_key != NAME:
                    self.logs.error(
                        f"Invalid row count comparison configuration key {configuration_key}", location=self.location
                    )
            self._pop_path_element()

        return RowCountComparisonCheckCfg(
            source_header=header_str,
            source_line=check_str,
            source_configurations=check_configurations,
            location=self.location,
            name=name,
            other_table_name=other_table_name,
            other_partition_name=other_partition_name,
            other_data_source_name=other_data_source_name,
        )

    def __parse_reference_check(
        self,
        antlr_reference_check,
        header_str: str,
        check_str: str,
        check_configurations: dict | None,
    ) -> CheckCfg:
        antlr_reference_check: SodaCLAntlrParser.Reference_checkContext = antlr_reference_check

        antlr_source_column_name_arg_list = antlr_reference_check.getTypedRuleContexts(
            SodaCLAntlrParser.Source_column_nameContext
        )
        source_column_names = [arg.getText() for arg in antlr_source_column_name_arg_list]

        target_table_name = antlr_reference_check.identifier().getText()

        antlr_target_column_name_arg_list = antlr_reference_check.getTypedRuleContexts(
            SodaCLAntlrParser.Target_column_nameContext
        )
        target_column_names = [arg.getText() for arg in antlr_target_column_name_arg_list]

        if len(source_column_names) == 0:
            self.logs.error(
                f"No source columns in reference check",
                location=self.location,
            )
        if len(target_column_names) == 0:
            self.logs.error(
                f"No target columns in reference check",
                location=self.location,
            )
        if len(source_column_names) != len(target_column_names):
            self.logs.error(
                f"Number of source and target column names must be equal",
                location=self.location,
            )

        name = None
        if isinstance(check_configurations, dict):
            self._push_path_element(check_str, check_configurations)
            name = self._get_optional(NAME, str)
            for configuration_key in check_configurations:
                if configuration_key != NAME:
                    self.logs.error(
                        f"Invalid row count comparison configuration key {configuration_key}", location=self.location
                    )
            self._pop_path_element()

        return ReferenceCheckCfg(
            source_header=header_str,
            source_line=check_str,
            source_configurations=check_configurations,
            location=self.location,
            name=name,
            source_column_names=source_column_names,
            target_table_name=target_table_name,
            target_column_names=target_column_names,
        )

    def __parse_freshness_check(
        self,
        antlr_freshness_check,
        header_str: str,
        check_str: str,
        check_configurations: dict | None,
    ) -> CheckCfg:
        antlr_freshness_check: SodaCLAntlrParser.Freshness_checkContext = antlr_freshness_check

        column_name = self.__antlr_parse_identifier(antlr_freshness_check.identifier())
        variable_name = "NOW"
        if antlr_freshness_check.freshness_variable():
            variable_name = self.__antlr_parse_identifier(antlr_freshness_check.freshness_variable().identifier())

        antlr_staleness_threshold = antlr_freshness_check.staleness_threshold()
        warn_staleness_threshold = None
        name = None
        if antlr_staleness_threshold:
            fail_staleness_threshold = self.parse_staleness_threshold(antlr_staleness_threshold.getText())
        else:
            self._push_path_element(check_str, check_configurations)
            fail_staleness_threshold_text = self._get_optional(FAIL, str)
            fail_staleness_threshold = self.parse_staleness_threshold_text(fail_staleness_threshold_text)
            warn_staleness_threshold_text = self._get_optional(WARN, str)
            warn_staleness_threshold = self.parse_staleness_threshold_text(warn_staleness_threshold_text)

            name = self._get_optional(NAME, str)
            for configuration_key in check_configurations:
                if configuration_key not in [NAME, WARN, FAIL]:
                    self.logs.error(
                        f"Invalid row count comparison configuration key {configuration_key}", location=self.location
                    )

            self._pop_path_element()

        return FreshnessCheckCfg(
            source_header=header_str,
            source_line=check_str,
            source_configurations=check_configurations,
            location=self.location,
            name=name,
            column_name=column_name,
            variable_name=variable_name,
            fail_staleness_threshold=fail_staleness_threshold,
            warn_staleness_threshold=warn_staleness_threshold,
        )

    def parse_staleness_threshold_text(self, staleness_threshold_text):
        if isinstance(staleness_threshold_text, str):
            if staleness_threshold_text.startswith("when > "):
                return self.parse_staleness_threshold(staleness_threshold_text[len("when > ") :])
            else:
                self.logs.error(
                    f'Invalid staleness threshold "{staleness_threshold_text}"',
                    location=self.location,
                )

    def parse_staleness_threshold(self, staleness_threshold_text: str) -> timedelta | None:
        try:
            if "d" in staleness_threshold_text:
                parts = staleness_threshold_text.split("d")
                days = int(parts[0])
                hours = int(parts[1]) if parts[1] != "" else 0
                return timedelta(days=days, hours=hours)
            elif "h" in staleness_threshold_text:
                parts = staleness_threshold_text.split("h")
                hours = int(parts[0])
                minutes = int(parts[1]) if parts[1] != "" else 0
                return timedelta(hours=hours, minutes=minutes)
            elif "m" in staleness_threshold_text:
                parts = staleness_threshold_text.split("m")
                minutes = int(parts[0])
                seconds = int(parts[1]) if parts[1] != "" else 0
                return timedelta(minutes=minutes, seconds=seconds)
        except:
            self.logs.error(
                f'Problem parsing freshness threshold "{staleness_threshold_text}"',
                location=self.location,
            )

    def __antlr_parse_threshold_condition(self, antlr_threshold) -> ThresholdCfg:
        antlr_comparator_threshold = antlr_threshold.comparator_threshold()
        if antlr_comparator_threshold:
            comparator = antlr_comparator_threshold.comparator().getText()
            antlr_threshold_value = antlr_comparator_threshold.threshold_value()
            threshold_value = self.__antlr_threshold_value(antlr_threshold_value)
            if comparator == "<":
                return ThresholdCfg(lt=threshold_value)
            if comparator == "<=":
                return ThresholdCfg(lte=threshold_value)
            if comparator == ">":
                return ThresholdCfg(gt=threshold_value)
            if comparator == ">=":
                return ThresholdCfg(gte=threshold_value)
            if comparator == "=":
                return ThresholdCfg(lte=threshold_value, gte=threshold_value)
            if comparator in ["!=", "<>"]:
                return ThresholdCfg(lt=threshold_value, gt=threshold_value, is_split_zone=True)
            self.logs.error(f"Unsupported comparator {comparator}", location=self.location)
            return None

        antlr_between_threshold: SodaCLAntlrParser.Between_thresholdContext = antlr_threshold.between_threshold()
        if antlr_between_threshold:
            lower_included = antlr_between_threshold.ROUND_LEFT() is None
            antlr_lower_value = antlr_between_threshold.threshold_value(0)
            lower_bound = self.__antlr_threshold_value(antlr_lower_value)
            upper_included = antlr_between_threshold.ROUND_RIGHT() is None
            antlr_upper_value = antlr_between_threshold.threshold_value(1)
            upper_bound = self.__antlr_threshold_value(antlr_upper_value)
            if lower_bound > upper_bound:
                self.logs.error(
                    f"Left lower bound should be less than the upper bound on the right {antlr_between_threshold.getText()}",
                    location=self.location,
                )
            threshold_condition_cfg = ThresholdCfg(
                lt=upper_bound if not upper_included else None,
                lte=upper_bound if upper_included else None,
                gt=lower_bound if not lower_included else None,
                gte=lower_bound if lower_included else None,
            )

            if antlr_between_threshold.NOT():
                threshold_condition_cfg = threshold_condition_cfg.get_inverse()

            return threshold_condition_cfg

        if antlr_threshold.anomaly_threshold():
            return AnomalyThresholdCfg()

        self.logs.error(f'Unknown threshold "{antlr_threshold.getText()}"', location=self.location)

    def __antlr_threshold_value(self, antlr_threshold_value):
        if antlr_threshold_value.signed_number():
            return self.__antlr_parse_signed_number(antlr_threshold_value.signed_number())
        if antlr_threshold_value.percentage():
            return self.__antlr_parse_signed_number(antlr_threshold_value.percentage())

    def __antlr_parse_signed_number(self, antlr_signed_number):
        signed_number_str = antlr_signed_number.getText()
        if signed_number_str.startswith("+"):
            signed_number_str = signed_number_str[1:]
        return float(signed_number_str)

    def __parse_missing_and_valid(
        self,
        configuration_type: str,
        configuration_value,
        missing_and_valid_cfg: MissingAndValidCfg,
    ) -> None:
        def set_configuration_value(value):
            field_name = configuration_type.replace(" ", "_")
            location_field_name = f"{field_name}_location"
            previous_location = getattr(missing_and_valid_cfg, location_field_name)
            if previous_location:
                self.logs.warning(
                    f"Overwriting {configuration_type} from {previous_location}",
                    location=self.location,
                )
            setattr(missing_and_valid_cfg, field_name, value)
            setattr(missing_and_valid_cfg, location_field_name, self.location)

        if configuration_type in ["valid values", "missing values"]:
            if all(isinstance(v, str) for v in configuration_value) or all(
                isinstance(v, Number) for v in configuration_value
            ):
                set_configuration_value(configuration_value)
            else:
                self.logs.error(
                    f"Only strings or only numbers allowed in {configuration_type}: {configuration_value}",
                    location=self.location,
                )

        elif configuration_type in [
            "valid min length",
            "valid max length",
            "valid length",
        ]:
            try:
                set_configuration_value(int(configuration_value))
            except ValueError:
                self.logs.error(
                    f"{configuration_type} must be an integer, but was {configuration_value}",
                    location=self.location,
                )

        elif configuration_type in ["valid min", "valid max"]:
            try:
                set_configuration_value(float(configuration_value))
            except ValueError:
                self.logs.error(
                    f"{configuration_type} must be an number (float), but was {configuration_value}",
                    location=self.location,
                )

        elif configuration_type in [
            "missing format",
            "missing regex",
            "valid regex",
            "valid format",
        ]:
            if isinstance(configuration_value, str):
                set_configuration_value(configuration_value)
            else:
                self.logs.error(
                    f"{configuration_type} must be a string, but was {type(configuration_value).__name__}",
                    location=self.location,
                )

    def __parse_table_filter_section(self, antlr_table_filter_header, header_str, header_content):
        from soda.sodacl.partition_cfg import PartitionCfg

        if isinstance(header_content, dict):
            table_name = self.__antlr_parse_identifier_name_from_header(antlr_table_filter_header)
            data_source_scan_cfg = self.sodacl_cfg._get_or_create_data_source_scan_cfgs(self.data_source_name)
            table_cfg: TableCfg = data_source_scan_cfg.get_or_create_table_cfg(table_name)

            partition_name = self.__antlr_parse_partition_from_header(antlr_table_filter_header)
            partition_cfg: PartitionCfg = table_cfg.create_partition(self.location.file_path, partition_name)
            partition_filter = header_content.get("where")
            if partition_filter is None:
                self.logs.error(
                    f'Expecting "where" with a SQL expression as a configuration under {header_str}',
                    location=self.location,
                )
            else:
                partition_cfg.sql_partition_filter = partition_filter

        else:
            self.logs.error(
                f'Skipping section "{header_str}" because content is not an object/dict',
                location=self.location,
            )

    def __parse_automated_monitoring_section(self, header_str, header_content):
        if isinstance(header_content, dict):
            automated_monitoring_cfg = AutomatedMonitoringCfg(self.data_source_name, self.location)
            data_source_scan_cfg = self.sodacl_cfg._get_or_create_data_source_scan_cfgs(self.data_source_name)
            data_source_scan_cfg.add_monitoring_cfg(automated_monitoring_cfg)
            automated_monitoring_cfg.data_source_name = header_content.get("data_source")
            tables = header_content.get("tables")
            if isinstance(tables, list):
                for table in tables:
                    if table.startswith("exclude "):
                        exclude_table_expression = table[len("exclude ") :]
                        automated_monitoring_cfg.exclude_tables.append(exclude_table_expression)
                    else:
                        if table.startswith("include "):
                            include_table_expression = table[len("include ") :]
                        else:
                            include_table_expression = table
                        automated_monitoring_cfg.include_tables.append(include_table_expression)
            else:
                self.logs.error(
                    'Content of "tables" must be a list of include and/or exclude expressions', location=self.location
                )

            find_anomalies = header_content.get("find anomalies")
            if find_anomalies:
                row_count = find_anomalies.get("row count")
                if isinstance(row_count, bool) and row_count == False:
                    automated_monitoring_cfg.row_count = False
                schema = find_anomalies.get("schema")
                if isinstance(schema, bool) and schema == False:
                    automated_monitoring_cfg.schema = False
        else:
            self.logs.error(
                f'Skipping section "{header_str}" because content is not an object/dict',
                location=self.location,
            )

    def __parse_discover_tables_section(self, header_str, header_content):
        if isinstance(header_content, dict):
            discover_tables_cfg = DiscoverTablesCfg(self.data_source_name, self.location)
            data_source_scan_cfg = self.sodacl_cfg._get_or_create_data_source_scan_cfgs(self.data_source_name)
            data_source_scan_cfg.add_discover_tables_cfg(discover_tables_cfg)
            discover_tables_cfg.data_source_name = header_content.get("data_source")
            tables = header_content.get("tables")
            if isinstance(tables, list):
                for table in tables:
                    if table.startswith("exclude "):
                        exclude_table_expression = table[len("exclude ") :]
                        discover_tables_cfg.exclude_tables.append(exclude_table_expression)
                    else:
                        if table.startswith("include "):
                            include_table_expression = table[len("include ") :]
                            discover_tables_cfg.include_tables.append(include_table_expression)
                        else:
                            include_table_expression = table
                            discover_tables_cfg.include_tables.append(include_table_expression)
            else:
                self.logs.error(
                    "Content of 'tables' must be a list of include and/or exclude expressions", location=self.location
                )
        else:
            self.logs.error(
                f'Skipping section "{header_str}" because content is not an object/dict',
                location=self.location,
            )

    def __parse_profile_columns_section(self, header_str, header_content):
        if isinstance(header_content, dict):
            profile_columns_cfg = ProfileColumnsCfg(self.data_source_name, self.location)
            data_source_scan_cfg = self.sodacl_cfg._get_or_create_data_source_scan_cfgs(self.data_source_name)
            data_source_scan_cfg.add_profile_columns_cfg(profile_columns_cfg)

            columns = header_content.get("columns")
            if isinstance(columns, list):
                for column_expression in columns:
                    if column_expression.startswith("exclude "):
                        exclude_column_expression = column_expression[len("exclude ") :]
                        profile_columns_cfg.exclude_columns.append(exclude_column_expression)
                    else:
                        if column_expression.startswith("include "):
                            include_column_expression = column_expression[len("include ") :]
                        else:
                            include_column_expression = column_expression
                        profile_columns_cfg.include_columns.append(include_column_expression)
            elif columns is None:
                self.logs.error('Configuration key "columns" is required in profile columns', location=self.location)
            else:
                self.logs.error('Content of "columns" must be a list of column expressions', location=self.location)
        else:
            self.logs.error(
                f'Skipping section "{header_str}" because content is not an object/dict',
                location=self.location,
            )

    def __parse_nameset_list(self, header_content, for_each_cfg):
        for name_filter_index, name_filter_str in enumerate(header_content):
            if isinstance(name_filter_str, str):
                is_include = True
                name_filter_pieces_str = name_filter_str
                if name_filter_pieces_str.startswith("include "):
                    name_filter_pieces_str = name_filter_pieces_str[len("include ") :]
                elif name_filter_pieces_str.startswith("exclude "):
                    name_filter_pieces_str = name_filter_pieces_str[len("exclude ") :]
                    is_include = False

                data_source_name_filter = None
                table_name_filter = None
                column_name_filter = None

                filter_pieces_list = re.split(r"\.", name_filter_pieces_str)
                if isinstance(for_each_cfg, ForEachTableCfg):
                    if len(filter_pieces_list) == 1:
                        data_source_name_filter = self.data_source_name
                        table_name_filter = filter_pieces_list[0]
                    elif len(filter_pieces_list) == 2:
                        data_source_name_filter = filter_pieces_list[0]
                        if data_source_name_filter != self.data_source_name:
                            self.logs.error(
                                f"Cross data source table sets not yet supported. {name_filter_pieces_str} "
                                f"refers to non default data source  {data_source_name_filter}. "
                                f"Default data source is {self.data_source_name}"
                            )
                            data_source_name_filter = None
                            table_name_filter = None
                        else:
                            table_name_filter = filter_pieces_list[1]
                    else:
                        self.logs.error(
                            f'Table name filter "{name_filter_str}" should exist out of 1 or 2 pieces separated by a dot.  Pieces count: {len(filter_pieces_list)}',
                            location=self.location,
                        )

                if isinstance(for_each_cfg, ForEachColumnCfg):
                    if len(filter_pieces_list) == 2:
                        data_source_name_filter = self.data_source_name
                        table_name_filter = filter_pieces_list[0]
                        column_name_filter = filter_pieces_list[1]
                    elif len(filter_pieces_list) == 3:
                        data_source_name_filter = filter_pieces_list[0]
                        if data_source_name_filter != self.data_source_name:
                            self.logs.error(
                                f"Cross data source column sets not yet supported. {name_filter_pieces_str} "
                                f"refers to non default data source  {data_source_name_filter}. "
                                f"Default data source is {self.data_source_name}"
                            )
                            data_source_name_filter = None
                            table_name_filter = None
                            column_name_filter = None
                        else:
                            table_name_filter = filter_pieces_list[1]
                            column_name_filter = filter_pieces_list[2]
                    else:
                        self.logs.error(
                            f'Column name filter "{name_filter_str}" should exist out of 2 or 3 pieces separated by a dot.  Pieces count: {len(filter_pieces_list)}',
                            location=self.location,
                        )

                if data_source_name_filter is not None:
                    name_filter = NameFilter(
                        data_source_name_filter=data_source_name_filter,
                        table_name_filter=table_name_filter,
                        partition_name_filter=None,
                        column_name_filter=column_name_filter,
                    )
                    if is_include:
                        for_each_cfg.includes.append(name_filter)
                    else:
                        for_each_cfg.excludes.append(name_filter)
            else:
                self.logs.error(
                    f'Name filter "{name_filter_str}" is not a string',
                    location=self.location,
                )

    def __antlr_parse_identifier_name_from_header(self, antlr_header, identifier_index: int = 0):
        antlr_identifier = antlr_header.getTypedRuleContext(SodaCLAntlrParser.IdentifierContext, identifier_index)
        return self.__antlr_parse_identifier(antlr_identifier) if antlr_identifier else None

    def __antlr_parse_partition_from_header(self, antlr_header):
        if antlr_header.partition_name():
            return self.__antlr_parse_identifier(antlr_header.partition_name().identifier())

    def __antlr_parse_identifier(self, antlr_identifier) -> str:
        return antlr_identifier.getText()
        # TODO consider resolving escape chars from a quoted strings:
        # identifier = re.sub(r'\\(.)', '\g<1>', unquoted_identifier)

    def __parse_antlr_checks_for_each_table_section(
        self, antlr_checks_for_each_table_header, header_str, header_content
    ):
        for_each_table_cfg = ForEachTableCfg()
        for_each_table_cfg.table_alias_name = self.__antlr_parse_identifier_name_from_header(
            antlr_checks_for_each_table_header
        )
        tables = self._get_required("tables", list)
        if tables:
            self._push_path_element("tables", tables)
            self.__parse_nameset_list(tables, for_each_table_cfg)
            self._pop_path_element()
        check_cfgs = self._get_required("checks", list)
        if check_cfgs:
            self._push_path_element("checks", check_cfgs)
            for_each_table_cfg.check_cfgs = self.__parse_checks_in_for_each_section(header_str, check_cfgs)
            self._pop_path_element()
        for_each_table_cfg.location = self.location
        self.sodacl_cfg.for_each_table_cfgs.append(for_each_table_cfg)

    def __parse_antlr_checks_for_each_column_section(
        self, antlr_checks_for_each_column_header, header_str, header_content
    ):
        # TODO atm, It is not necessary that the column variable names in the metrics match.  So they are discarded for now.
        #  That's because *all* column references in the checks will be replaced anyways.
        for_each_column_cfg = ForEachColumnCfg()
        self.__parse_nameset_list(header_content, for_each_column_cfg)
        for_each_column_cfg.table_alias_name = self.__antlr_parse_identifier_name_from_header(
            antlr_checks_for_each_column_header
        )
        check_cfgs = self._get_required("checks", dict)
        for_each_column_cfg.check_cfgs = self.__parse_checks_in_for_each_section(header_str, check_cfgs)
        for_each_column_cfg.location = self.location
        self.sodacl_cfg.for_each_column_cfgs.append(for_each_column_cfg)

    def __parse_checks_in_for_each_section(self, header_str, header_content):
        check_cfgs: list[CheckCfg] = []
        if isinstance(header_content, list):
            for check_index, check_list_element in enumerate(header_content):
                self._push_path_element(check_index, check_list_element)

                check_str: str = None
                check_configurations = None

                if isinstance(check_list_element, str):
                    check_str = check_list_element
                elif isinstance(check_list_element, dict):
                    check_str = next(iter(check_list_element))
                    check_configurations = check_list_element[check_str]
                else:
                    self.logs.error(
                        f"Skipping unsupported check definition: {to_yaml_str(check_list_element)}",
                        location=self.location,
                    )

                if check_str is not None:
                    check_cfg = self.__parse_table_check_str(header_str, check_str, check_configurations)
                    check_cfgs.append(check_cfg)

                self._pop_path_element()
        else:
            self.logs.error(
                f'Skipping section "{header_str}" because content is not a list',
                location=self.location,
            )
        return check_cfgs

    def antlr_parse_check(self, text: str) -> AntlrParser:
        return AntlrParser(text, lambda p: p.check())

    def antlr_parse_section_header(self, text: str) -> AntlrParser:
        return AntlrParser(text, lambda p: p.section_header())

    def antlr_parse_column_configuration(self, text: str) -> AntlrParser:
        return AntlrParser(text, lambda p: p.configuration())

    def antlr_parse_threshold(self, text: str) -> AntlrParser:
        return AntlrParser(text, lambda p: p.threshold())


class AntlrParser(ErrorListener):
    def __init__(self, text: str, parser_function):
        self.text = text
        self.error_message = None
        self.exception = None

        input_stream = InputStream(text)
        lexer = SodaCLAntlrLexer(input_stream)
        lexer.removeErrorListeners()
        lexer.addErrorListener(self)
        stream = CommonTokenStream(lexer)
        parser = SodaCLAntlrParser(stream)
        parser.removeErrorListeners()
        parser.addErrorListener(self)

        self.result = parser_function(parser)

    def is_ok(self):
        return self.result is not None and self.error_message is None and self.exception is None

    def get_error_message(self):
        return self.error_message

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        self.error_message = msg
        self.exception = e



from __future__ import annotations

from soda.sodacl.column_configurations_cfg import ColumnConfigurationsCfg
from soda.sodacl.location import Location
from soda.sodacl.partition_cfg import PartitionCfg


class TableCfg:
    def __init__(self, table_name: str):
        self.table_name: str = table_name
        self.column_configurations_cfgs: dict[str, ColumnConfigurationsCfg] = {}
        self.column_configuration_locations: list[Location] = []
        self.partition_cfgs: list[PartitionCfg] = []

    def get_or_create_column_configurations(self, column_name: str) -> ColumnConfigurationsCfg:
        column_configurations_cfg = self.column_configurations_cfgs.get(column_name)
        if not column_configurations_cfg:
            column_configurations_cfg = ColumnConfigurationsCfg(column_name)
            self.column_configurations_cfgs[column_name] = column_configurations_cfg
        return column_configurations_cfg

    def create_partition(self, file_path: str, partition_name: str) -> PartitionCfg:
        # Because the default table level checks are modelled with filter_name = None
        if partition_name is None:
            # file paths should be ignored
            file_path = None
        partition_cfg = PartitionCfg(file_path=file_path, partition_name=partition_name)
        self.partition_cfgs.append(partition_cfg)
        return partition_cfg

    def find_partition(self, file_path: str, partition_name: str) -> PartitionCfg:
        # Because the default table level checks are modelled with filter_name = None
        if partition_name is None:
            # file paths should be ignored
            file_path = None
        partition_cfg = next(
            (
                filter
                for filter in self.partition_cfgs
                if filter.file_path == file_path and filter.partition_name == partition_name
            ),
            None,
        )
        # Because the default table level checks are modelled with filter_name = None
        if partition_cfg is None and partition_name is None:
            return self.create_partition(None, None)
        return partition_cfg



import logging
from typing import Tuple

from soda.sampler.sample_context import SampleContext
from soda.sampler.sample_ref import SampleRef
from soda.sampler.sampler import Sampler

logger = logging.getLogger(__name__)


class LogSampler(Sampler):
    def store_sample(self, sample_context: SampleContext) -> SampleRef:
        table_text, column_count, row_count = self.pretty_print(sample_context)
        sample_name = sample_context.sample_name
        sample_context.logs.info(f"Sample {sample_name}:\n{table_text}")
        return SampleRef(
            name=sample_name,
            schema=sample_context.sample.get_schema(),
            total_row_count=row_count,
            stored_row_count=row_count,
            type="log",
            message=f'Search in the console for "Sample {sample_name}"',
        )

    def pretty_print(self, sample_context: SampleContext, max_column_length: int = 25) -> Tuple[str, int, int]:
        def stringify(value, quote_strings):
            if isinstance(value, str):
                return f"'{value}'" if quote_strings else value
            if value is None:
                return ""
            return str(value)

        def maxify(value):
            return value if len(value) <= max_column_length else f"{value[:max_column_length - 3]}..."

        def serialize_row(row, quote_strings: bool = False):
            return [maxify(stringify(value, quote_strings)) for value in row]

        names = []
        lengths = []
        rules = []

        rows = sample_context.sample.get_rows()
        rows = [serialize_row(row, quote_strings=True) for row in rows]

        sample_columns = sample_context.sample.get_schema().columns
        column_names = serialize_row([sample_column.name for sample_column in sample_columns])
        for column_name in column_names:
            names.append(column_name)
            lengths.append(len(column_name))
        for column_index in range(len(lengths)):
            rls = [len(row[column_index]) for row in rows if row[column_index]]
            lengths[column_index] = max([lengths[column_index]] + rls)
            rules.append("-" * lengths[column_index])
        format = " ".join(["%%-%ss" % l for l in lengths])
        result = [format % tuple(names)]
        result.append(format % tuple(rules))
        for row in rows:
            result.append(format % tuple(row))
        return "\n".join(result), len(names), len(rows)



class ChangeOverTimeCfg:
    def __init__(self):
        self.last_measurements: Optional[int] = None
        self.last_aggregation: Optional[str] = None
        self.same_day_last_week: bool = False
        self.same_day_last_month: bool = False

    def to_jsonnable(self):
        jsonnable = {}
        if self.last_measurements:
            jsonnable["last_measurements"] = self.last_measurements
        if self.last_aggregation:
            jsonnable["last_aggregation"] = self.last_aggregation
        if self.same_day_last_week:
            jsonnable["same_day_last_week"] = self.same_day_last_week
        if self.same_day_last_month:
            jsonnable["same_day_last_month"] = self.same_day_last_month
        return jsonnable



from typing import List, Optional

from soda.sodacl.location import Location
from soda.sodacl.metric_check_cfg import MetricCheckCfg
from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg
from soda.sodacl.threshold_cfg import ThresholdCfg


class AnomalyMetricCheckCfg(MetricCheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: Optional[str],
        location: Location,
        name: Optional[str],
        metric_name: str,
        metric_args: Optional[List[object]],
        missing_and_valid_cfg: Optional[MissingAndValidCfg],
        filter: Optional[str],
        condition: Optional[str],
        metric_expression: Optional[str],
        metric_query: Optional[str],
        change_over_time_cfg: Optional["ChangeOverTimeCfg"],
        fail_threshold_cfg: Optional[ThresholdCfg],
        warn_threshold_cfg: Optional[ThresholdCfg],
        is_automated_monitoring: bool = False,
    ):
        super().__init__(
            source_header,
            source_line,
            source_configurations,
            location,
            name,
            metric_name,
            metric_args,
            missing_and_valid_cfg,
            filter,
            condition,
            metric_expression,
            metric_query,
            change_over_time_cfg,
            fail_threshold_cfg,
            warn_threshold_cfg,
        )
        self.is_automated_monitoring = is_automated_monitoring



from typing import List, Optional

from soda.sodacl.location import Location


class AutomatedMonitoringCfg:
    def __init__(self, data_source_name: str, location: Location):
        self.data_source_name: str = data_source_name
        self.location: Location = location
        self.include_tables: Optional[List[str]] = []
        self.exclude_tables: Optional[List[str]] = []
        self.row_count: bool = True
        self.schema: bool = True
        self.min_history = 3
        self.max_history = 90



from typing import List

from soda.profiling.discover_tables_result_column import DiscoverTablesResultColumn


class DiscoverTablesResultTable:
    def __init__(self, table_name: str, data_source: str, row_count: int):
        self.table_name: str = table_name
        self.data_source: str = data_source
        self.row_count: int = row_count
        self.result_columns: List[DiscoverTablesResultColumn] = []

    def create_column(self, column_name: str, column_type: str) -> DiscoverTablesResultColumn:
        column = DiscoverTablesResultColumn(column_name, column_type)
        self.result_columns.append(column)
        return column

    def get_cloud_dict(self) -> dict:
        cloud_dict = {
            "table": self.table_name,
            "dataSource": self.data_source,
            "rowCount": self.row_count,
            "schema": [result_column.get_cloud_dict() for result_column in self.result_columns],
        }
        return cloud_dict



from soda.execution.query import Query


class ReferenceQuery(Query):
    @staticmethod
    def build_source_column_list(metric):
        source_column_names = metric.check.check_cfg.source_column_names
        return ",".join(source_column_names)

    def __init__(
        self,
        data_source_scan: "DataSourceScan",
        metric: "ReferentialIntegrityMetric",
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            unqualified_query_name=f"reference[{ReferenceQuery.build_source_column_list(metric)}]",
        )

        from soda.sodacl.reference_check_cfg import ReferenceCheckCfg

        check_cfg: ReferenceCheckCfg = metric.check.check_cfg
        source_table_name = metric.partition.table.table_name
        source_column_names = check_cfg.source_column_names
        target_table_name = check_cfg.target_table_name
        target_column_names = check_cfg.target_column_names

        source_diagnostic_column_fields = "SOURCE.*"

        # TODO add global config of table diagnostic columns and apply that here
        # source_diagnostic_column_names = check_cfg.source_diagnostic_column_names
        # if source_diagnostic_column_names:
        #     source_diagnostic_column_names += source_column_names
        #     source_diagnostic_column_fields = ', '.join([f'SOURCE.{column_name}' for column_name in source_diagnostic_column_names])

        join_condition = " AND ".join(
            [
                f"SOURCE.{source_column_name} = TARGET.{target_column_names[index]}"
                for index, source_column_name in enumerate(source_column_names)
            ]
        )
        where_condition = " OR ".join(
            [f"TARGET.{target_column_name} IS NULL" for target_column_name in target_column_names]
        )

        self.sql = (
            f"SELECT {source_diagnostic_column_fields} \n"
            f"FROM {source_table_name} as SOURCE \n"
            f"     LEFT JOIN {target_table_name} as TARGET on {join_condition} \n"
            f"WHERE {where_condition}"
        )
        self.metric = metric

    def execute(self):
        self.store()
        if self.sample_ref:
            self.metric.value = self.sample_ref.total_row_count
            self.metric.invalid_references_sample_ref = self.sample_ref



from typing import Dict, List, Optional

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.column_checks_cfg import ColumnChecksCfg
from soda.sodacl.location import Location


class PartitionCfg:
    def __init__(self, file_path: str, partition_name: Optional[str] = None):
        self.file_path: str = file_path
        # partition_name None means the default partition (without filter)
        self.partition_name: Optional[str] = partition_name
        self.sql_partition_filter: Optional[str] = None
        self.check_cfgs: List[CheckCfg] = []
        self.column_checks_cfgs: Dict[str, ColumnChecksCfg] = {}
        self.locations: List[Location] = []

    def add_check_cfg(self, check_cfg: CheckCfg):
        self.check_cfgs.append(check_cfg)

    def get_or_create_column_checks(self, column_name: str) -> ColumnChecksCfg:
        column_checks_cfg = self.column_checks_cfgs.get(column_name)
        if not column_checks_cfg:
            column_checks_cfg = ColumnChecksCfg(column_name)
            self.column_checks_cfgs[column_name] = column_checks_cfg
        return column_checks_cfg



from __future__ import annotations

import datetime
import hashlib
import importlib
import json
import re
from datetime import date, datetime
from math import ceil, floor
from numbers import Number
from textwrap import dedent

from soda.common.exceptions import DataSourceError
from soda.common.logs import Logs
from soda.execution.data_type import DataType
from soda.execution.partition_queries import PartitionQueries
from soda.execution.query import Query
from soda.telemetry.soda_telemetry import SodaTelemetry

soda_telemetry = SodaTelemetry.get_instance()


class DataSource:

    """
    Implementing a DataSource:
    @m1n0, can you add a checklist here of places where DataSource implementors need to make updates to add
    a new DataSource?

    Validation of the connection configuration properties:
    The DataSource impl is only responsible to raise an exception with an appropriate message in te #connect()
    See that abstract method below for more details.
    """

    # Maps synonym types for the convenience of use in checks.
    # Keys represent the data_source type, values are lists of "aliases" that can be used in SodaCL as synonyms.
    SCHEMA_CHECK_TYPES_MAPPING: dict = {
        "character varying": ["varchar"],
    }
    SQL_TYPE_FOR_CREATE_TABLE_MAP: dict = {
        DataType.TEXT: "VARCHAR(255)",
        DataType.INTEGER: "INT",
        DataType.DECIMAL: "FLOAT",
        DataType.DATE: "DATE",
        DataType.TIME: "TIME",
        DataType.TIMESTAMP: "TIMESTAMP",
        DataType.TIMESTAMP_TZ: "TIMESTAMPTZ",
        DataType.BOOLEAN: "BOOLEAN",
    }

    SQL_TYPE_FOR_SCHEMA_CHECK_MAP = {
        DataType.TEXT: "character varying",
        DataType.INTEGER: "integer",
        DataType.DECIMAL: "double precision",
        DataType.DATE: "date",
        DataType.TIME: "time",
        DataType.TIMESTAMP: "timestamp without time zone",
        DataType.TIMESTAMP_TZ: "timestamp with time zone",
        DataType.BOOLEAN: "boolean",
    }

    NUMERIC_TYPES_FOR_PROFILING = ["integer", "double precision"]
    TEXT_TYPES_FOR_PROFILING = ["character varying"]

    @staticmethod
    def create(
        logs: Logs,
        data_source_name: str,
        connection_type: str,
        data_source_properties: dict,
        connection_properties: dict,
    ) -> DataSource:
        """
        The returned data_source does not have a connection.  It is the responsibility of
        the caller to initialize data_source.connection.  To create a new connection,
        use data_source.connect(...)
        """
        module_name = f"soda.data_sources.{connection_type}_data_source"
        data_source_properties["connection_type"] = connection_type
        try:
            module = importlib.import_module(module_name)
            return module.DataSourceImpl(logs, data_source_name, data_source_properties, connection_properties)
        except ModuleNotFoundError as e:
            if connection_type == "postgresql":
                logs.error(f'Data source type "{connection_type}" not found. Did you mean postgres?')
            else:
                raise DataSourceError(
                    f'Data source type "{connection_type}" not found. Did you spell {connection_type} correct? Did you install module soda-core-{connection_type}?'
                )
            return None

    def __init__(self, logs: Logs, data_source_name: str, data_source_properties: dict, connection_properties: dict):
        self.logs = logs
        self.data_source_name = data_source_name
        self.data_source_properties: dict = data_source_properties
        self.connection_properties = connection_properties
        # Pep 249 compliant connection object (aka DBAPI)
        # https://www.python.org/dev/peps/pep-0249/#connection-objects
        # @see self.connect() for initialization
        self.type = self.data_source_properties.get("connection_type")
        self.connection = None
        self.database: str = data_source_properties.get("database")
        self.schema: str | None = data_source_properties.get("schema")
        self.table_prefix = data_source_properties.get("table_prefix")

        soda_telemetry.set_attribute("datasource_type", self.type)
        # TODO get hash after init so that children can implement init
        soda_telemetry.set_attribute("datasource_id", soda_telemetry.obtain_datasource_hash(self))

    def validate_configuration(self, connection_properties: dict, logs: Logs) -> None:
        """
        validates connection_properties and self.data_source_properties
        """
        raise NotImplementedError(f"TODO: Implement {type(self)}.validate_configuration(...)")

    def get_type_name(self, type_code):
        return str(type_code)

    def create_partition_queries(self, partition):
        return PartitionQueries(partition)

    def is_supported_metric_name(self, metric_name: str) -> bool:
        return (
            metric_name in ["row_count", "missing_count", "invalid_count", "valid_count", "duplicate_count"]
            or self.get_metric_sql_aggregation_expression(metric_name, None, None) is not None
        )

    def get_metric_sql_aggregation_expression(self, metric_name: str, metric_args: list[object] | None, expr: str):
        if "min" == metric_name:
            return self.expr_min(expr)
        if "max" == metric_name:
            return self.expr_max(expr)
        if "avg" == metric_name:
            return self.expr_avg(expr)
        if "sum" == metric_name:
            return self.expr_sum(expr)
        if "min_length" == metric_name:
            return self.expr_min(self.expr_length(expr))
        if "max_length" == metric_name:
            return self.expr_max(self.expr_length(expr))
        if "avg_length" == metric_name:
            return self.expr_avg(self.expr_length(expr))
        return None

    def is_same_type_in_schema_check(self, expected_type: str, actual_type: str):
        expected_type = expected_type.strip().lower()

        if (
            actual_type in self.SCHEMA_CHECK_TYPES_MAPPING
            and expected_type in self.SCHEMA_CHECK_TYPES_MAPPING[actual_type]
        ):
            return True

        return expected_type == actual_type

    ######################
    # SQL Queries
    ######################

    def sql_to_get_column_metadata_for_table(self, table_name: str) -> str:
        sql = (
            f"SELECT column_name, data_type, is_nullable \n"
            f"FROM information_schema.columns \n"
            f"WHERE lower(table_name) = '{table_name.lower()}'"
        )
        if self.database:
            sql += f" \n  AND lower(table_catalog) = '{self.database.lower()}'"
        if self.schema:
            sql += f" \n  AND lower(table_schema) = '{self.schema.lower()}'"

        sql += f"\nORDER BY ORDINAL_POSITION"

        return sql

    def sql_get_table_names_with_count(
        self, include_tables: list[str] | None = None, exclude_tables: list[str] | None = None
    ) -> str:
        table_filter_expression = self.sql_table_include_exclude_filter(
            "relname", "schemaname", include_tables, exclude_tables
        )
        where_clause = f"\nWHERE {table_filter_expression} \n" if table_filter_expression else ""
        return f"SELECT relname, n_live_tup \n" f"FROM pg_stat_user_tables" f"{where_clause}"

    def sql_get_column(self, include_tables: list[str] | None = None, exclude_tables: list[str] | None = None) -> str:
        table_filter_expression = self.sql_table_include_exclude_filter(
            "table_name", "table_schema", include_tables, exclude_tables
        )
        where_clause = f"\nWHERE {table_filter_expression} \n" if table_filter_expression else ""
        return (
            f"SELECT table_name, column_name, data_type, is_nullable \n"
            f"FROM information_schema.columns"
            f"{where_clause}"
        )

    def sql_get_table_count(self, table_name: str) -> str:
        return f"SELECT count(*) from {table_name}"

    def sql_table_include_exclude_filter(
        self,
        table_column_name: str,
        schema_column_name: str | None = None,
        include_tables: list[str] = [],
        exclude_tables: list[str] = [],
    ) -> str | None:
        tablename_filter_clauses = []
        if include_tables:
            sql_include_clauses = " OR ".join(
                [f"lower({table_column_name}) like '{include_table.lower()}'" for include_table in include_tables]
            )
            tablename_filter_clauses.append(f"({sql_include_clauses})")

        if exclude_tables:
            tablename_filter_clauses.extend(
                [f"lower({table_column_name}) not like '{exclude_table.lower()}'" for exclude_table in exclude_tables]
            )

        if hasattr(self, "schema") and self.schema and schema_column_name:
            tablename_filter_clauses.append(f"lower({schema_column_name}) = '{self.schema.lower()}'")
        return "\n      AND ".join(tablename_filter_clauses) if tablename_filter_clauses else None

    def sql_find_table_names(
        self,
        filter: str | None = None,
        include_tables: list[str] = [],
        exclude_tables: list[str] = [],
        table_column_name: str = "table_name",
        schema_column_name: str = "table_schema",
    ) -> str:
        sql = f"SELECT table_name \n" f"FROM {self.sql_information_schema_identifier()}"
        where_clauses = []

        if filter:
            where_clauses.append(f"lower({table_column_name}) like '{filter.lower()}'")

        includes_excludes_filter = self.sql_table_include_exclude_filter(
            table_column_name, schema_column_name, include_tables, exclude_tables
        )
        if includes_excludes_filter:
            where_clauses.append(includes_excludes_filter)

        if where_clauses:
            where_clauses_sql = "\n  AND ".join(where_clauses)
            sql += f"\nWHERE {where_clauses_sql}"

        return sql

    def sql_information_schema_identifier(self) -> str:
        return "information_schema.tables"

    def sql_analyze_table(self, table: str) -> str | None:
        return None

    def profiling_sql_values_frequencies_query(self, table_name: str, column_name: str) -> str:
        return dedent(
            f"""
                WITH freq_values AS (
                  {self.profiling_sql_cte_value_frequencies(table_name, column_name)}
                )
                {self.profiling_sql_value_frequencies_select()}
            """
        )

    def profiling_sql_top_values(self, table_name: str, column_name: str) -> str:
        return dedent(
            f"""
                WITH freq_values AS (
                  {self.profiling_sql_cte_value_frequencies(table_name, column_name)}
                )
                {self.profiling_sql_frequent_values_cte('freq_values', 'frequent_values', is_final=True)}
            """
        )

    def profiling_sql_cte_value_frequencies(self, table_name: str, column_name: str) -> str:
        column_name = self.quote_column(column_name)
        return dedent(
            f"""
                SELECT {column_name} as value_name, count(*) as frequency
                FROM {table_name}
                GROUP BY value_name
            """
        )

    def profiling_sql_frequent_values_cte(self, source_table_name: str, cte_name: str, is_final: bool = False) -> str:
        sql = dedent(
            f"""
            , {cte_name} as (
                SELECT
                    frequency
                    , row_number() over (order by frequency desc) as idx
                    , value_name
                FROM {source_table_name}
                ORDER BY frequency desc
                LIMIT 5
            )
            """
        )
        if is_final:
            sql += f"\n SELECT * FROM {cte_name}"
        return sql

    def profiling_sql_value_frequencies_select(self) -> str:
        return dedent(
            f"""
            , mins as (
            SELECT value_name, ROW_NUMBER() OVER(order by value_name asc) as idx, frequency, 'mins' as metric_name
            FROM freq_values
            WHERE value_name is not null
            ORDER BY value_name asc
            LIMIT 5
        )
        , maxes as (

            SELECT value_name, ROW_NUMBER() OVER(order by value_name desc) as idx, frequency, 'maxes' as metric_name
            FROM freq_values
            WHERE value_name is not null
            ORDER BY value_name desc
            LIMIT 5
        )
        {self.profiling_sql_frequent_values_cte(source_table_name='freq_values', cte_name='frequent_values', is_final=False)}
        , final as (
            SELECT
                mins.value_name as mins
                , maxes.value_name as maxes
                , frequent_values.value_name as frequent_values
                , frequent_values.frequency as frequency
            FROM mins
            JOIN maxes
                 on mins.idx = maxes.idx
            JOIN frequent_values
                on mins.idx = frequent_values.idx
            ORDER BY frequency desc
        )
        SELECT * FROM final
            """
        )

    def profiling_sql_numeric_aggregates(self, table_name: str, column_name: str) -> str:
        column_name = self.quote_column(column_name)
        return dedent(
            f"""
            SELECT
                avg({column_name}) as average
                , sum({column_name}) as sum
                , variance({column_name}) as variance
                , stddev({column_name}) as standard_deviation
                , count(distinct({column_name})) as distinct_values
                , sum(case when {column_name} is null then 1 else 0 end) as missing_values
            FROM {table_name}
            """
        )

    def profiling_sql_text_aggregates(self, table_name: str, column_name: str) -> str:
        column_name = self.quote_column(column_name)
        return dedent(
            f"""
            SELECT
                count(distinct({column_name})) as distinct_values
                , sum(case when {column_name} is null then 1 else 0 end) as missing_values
                , avg(length({column_name})) as avg_length
                , min(length({column_name})) as min_length
                , max(length({column_name})) as max_length
            FROM {table_name}
            """
        )

    def histogram_sql_and_boundaries(
        self, table_name: str, column_name: str, min: int | float, max: int | float
    ) -> tuple[str, list[int | float]]:
        # TODO: make configurable or derive dynamically based on data quantiles etc.
        number_of_bins: int = 20

        if not min < max:
            self.logs.error(
                f"Min of {column_name} on table: {table_name} must be smaller than max value. Min is {min}, and max is {max}"
            )
            return "", []

        min_value = floor(min * 1000) / 1000
        max_value = ceil(max * 1000) / 1000
        bin_width = (max_value - min_value) / number_of_bins

        boundary_start = min_value
        bins_list = [min_value]
        for _ in range(0, number_of_bins):
            boundary_start += bin_width
            bins_list.append(round(boundary_start, 3))

        field_clauses = []
        for i in range(0, number_of_bins):
            lower_bound = "" if i == 0 else f"{bins_list[i]} <= value_name"
            upper_bound = "" if i == number_of_bins - 1 else f"value_name < {bins_list[i+1]}"
            optional_and = "" if lower_bound == "" or upper_bound == "" else " and "
            field_clauses.append(f"SUM(CASE WHEN {lower_bound}{optional_and}{upper_bound} then frequency END)")

        fields = ",\n ".join(field_clauses)

        sql = (
            f"WITH freq_values as ({self.profiling_sql_cte_value_frequencies(table_name, column_name)})\n"
            f"SELECT {fields} From freq_values"
        )
        return sql, bins_list

    ######################
    # Query Execution
    ######################

    def get_row_counts_all_tables(
        self,
        include_tables: list[str] | None = None,
        exclude_tables: list[str] | None = None,
        query_name: str | None = None,
    ) -> dict[str, int]:
        """
        Returns a dict that maps table names to row counts.
        """
        sql = self.sql_get_table_names_with_count(include_tables=include_tables, exclude_tables=exclude_tables)
        if sql:
            query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=query_name or "get_row_counts_all_tables",
                sql=sql,
            )
            query.execute()
            return {row[0]: row[1] for row in query.rows}
        # Single query to get the metadata not available, get the counts one by one.
        all_tables = self.get_table_names(include_tables=include_tables, exclude_tables=exclude_tables)
        result = {}

        for table in all_tables:
            query_name_str = f"get_row_count_{table}"
            if query_name:
                query_name_str = f"{query_name}_{table}"
            query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=query_name_str,
                sql=self.sql_get_table_count(self.quote_table(table)),
            )
            query.execute()
            if query.rows:
                result[table] = query.rows[0][0]

        return result

    def get_table_names(
        self,
        filter: str | None = None,
        include_tables: list[str] = [],
        exclude_tables: list[str] = [],
        query_name: str | None = None,
    ) -> list[str]:
        sql = self.sql_find_table_names(filter, include_tables, exclude_tables)
        query = Query(
            data_source_scan=self.data_source_scan, unqualified_query_name=query_name or "get_table_names", sql=sql
        )
        query.execute()
        table_names = [row[0] for row in query.rows]
        return table_names

    def analyze_table(self, table: str):
        if self.sql_analyze_table(table):
            Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"analyze_{table}",
                sql=self.sql_analyze_table(table),
            ).execute()

    def quote_table_declaration(self, table_name) -> str:
        return self.quote_table(table_name=table_name)

    def quote_table(self, table_name) -> str:
        return f'"{table_name}"'

    def prefix_table(self, table_name: str) -> str:
        if self.table_prefix:
            return f"{self.table_prefix}.{table_name}"
        return table_name

    def quote_column_declaration(self, column_name: str) -> str:
        return self.quote_column(column_name)

    def quote_column(self, column_name: str) -> str:
        return f'"{column_name}"'

    def get_sql_type_for_create_table(self, data_type: str) -> str:
        if data_type in self.SQL_TYPE_FOR_CREATE_TABLE_MAP:
            return self.SQL_TYPE_FOR_CREATE_TABLE_MAP.get(data_type)
        else:
            return data_type

    def get_sql_type_for_schema_check(self, data_type: str) -> str:
        data_source_type = self.SQL_TYPE_FOR_SCHEMA_CHECK_MAP.get(data_type)
        if data_source_type is None:
            raise NotImplementedError(
                f"Data type {data_type} is not mapped in {type(self)}.SQL_TYPE_FOR_SCHEMA_CHECK_MAP"
            )
        return data_source_type

    def literal(self, o: object):
        if o is None:
            return "NULL"
        elif isinstance(o, Number):
            return self.literal_number(o)
        elif isinstance(o, str):
            return self.literal_string(o)
        elif isinstance(o, datetime):
            return self.literal_datetime(o)
        elif isinstance(o, date):
            return self.literal_date(o)
        elif isinstance(o, list) or isinstance(o, set) or isinstance(o, tuple):
            return self.literal_list(o)
        elif isinstance(o, bool):
            return self.literal_boolean(o)
        raise RuntimeError(f"Cannot convert type {type(o)} to a SQL literal: {o}")

    def literal_number(self, value: Number):
        if value is None:
            return None
        return str(value)

    def literal_string(self, value: str):
        if value is None:
            return None
        return "'" + self.escape_string(value) + "'"

    def literal_list(self, l: list):
        if l is None:
            return None
        return "(" + (",".join([self.literal(e) for e in l])) + ")"

    def literal_date(self, date: date):
        date_string = date.strftime("%Y-%m-%d")
        return f"DATE '{date_string}'"

    def literal_datetime(self, datetime: datetime):
        return f"'{datetime.isoformat()}'"

    def literal_boolean(self, boolean: bool):
        return "TRUE" if boolean is True else "FALSE"

    def expr_count_all(self) -> str:
        return "COUNT(*)"

    def expr_count_conditional(self, condition: str):
        return f"COUNT(CASE WHEN {condition} THEN 1 END)"

    def expr_conditional(self, condition: str, expr: str):
        return f"CASE WHEN {condition} THEN {expr} END"

    def expr_count(self, expr):
        return f"COUNT({expr})"

    def expr_distinct(self, expr):
        return f"DISTINCT({expr})"

    def expr_length(self, expr):
        return f"LENGTH({expr})"

    def expr_min(self, expr):
        return f"MIN({expr})"

    def expr_max(self, expr):
        return f"MAX({expr})"

    def expr_avg(self, expr):
        return f"AVG({expr})"

    def expr_sum(self, expr):
        return f"SUM({expr})"

    def expr_regexp_like(self, expr: str, regex_pattern: str):
        return f"REGEXP_LIKE({expr}, '{regex_pattern}')"

    def expr_in(self, left: str, right: str):
        return f"{left} IN {right}"

    def cast_text_to_number(self, column_name, validity_format: str):
        regex = self.escape_regex(r"'[^-0-9\.\,]'")
        return f"CAST(NULLIF(REGEXP_REPLACE(REGEXP_REPLACE({column_name}, {regex}, ''{self.regex_replace_flags()}), ',', '.'{self.regex_replace_flags()}), '') AS {self.SQL_TYPE_FOR_CREATE_TABLE_MAP[DataType.DECIMAL]})"

    def regex_replace_flags(self) -> str:
        return ", 'g'"

    def escape_string(self, value: str):
        return re.sub(r"(\\.)", r"\\\1", value)

    def escape_regex(self, value: str):
        return value

    def get_max_aggregation_fields(self):
        """
        Max number of fields to be aggregated in 1 aggregation query
        """
        return 50

    def connect(self, connection_properties: dict):
        """
        Subclasses use self.connection_properties to initialize self.connection with a PEP 249 connection

        Any BaseException may be raised in case of errors in the connection_properties or in case
        the database.connect itself fails for some other reason.  The caller of this method will
        catch the exception and add an error log to the scan.
        """
        raise NotImplementedError(f"TODO: Implement {type(self)}.connect()")

    def fetchall(self, sql: str):
        # TODO: Deprecated - not used, use Query object instead.
        try:
            cursor = self.connection.cursor()
            try:
                self.logs.info(f"Query: \n{sql}")
                cursor.execute(sql)
                return cursor.fetchall()
            finally:
                cursor.close()
        except BaseException as e:
            self.logs.error(f"Query error: {e}\n{sql}", exception=e)
            self.query_failed(e)

    def is_connected(self):
        return self.connection is not None

    def disconnect(self) -> None:
        if self.connection:
            self.connection.close()
            # self.connection = None is used in self.is_connected
            self.connection = None

    def commit(self):
        self.connection.commit()

    def query_failed(self, e: BaseException):
        self.rollback()

    def rollback(self):
        self.connection.rollback()

    def create_data_source_scan(self, scan: Scan, data_source_scan_cfg: DataSourceScanCfg):
        from soda.execution.data_source_scan import DataSourceScan

        data_source_scan = DataSourceScan(scan, data_source_scan_cfg, self)
        self.data_source_scan = data_source_scan

        return self.data_source_scan

    @staticmethod
    def format_column_default(identifier: str) -> str:
        """Formats column identifier to e.g. a default case for a given data source."""
        return identifier

    @staticmethod
    def format_type_default(identifier: str) -> str:
        """Formats type identifier to e.g. a default case for a given data source."""
        return identifier

    def safe_connection_data(self):
        """Return non-critically sensitive connection details.

        Useful for debugging.
        """
        # to be overridden by subclass

    def generate_hash_safe(self):
        """Generates a safe hash from non-sensitive connection details.

        Useful for debugging, identifying data sources anonymously and tracing.
        """
        data = self.safe_connection_data()

        return self.hash_data(data)

    def hash_data(self, data) -> str:
        """Hash provided data using a non-reversible hashing algorithm."""
        encoded = json.dumps(data, sort_keys=True).encode()
        return hashlib.sha256(encoded).hexdigest()



from __future__ import annotations

from datetime import datetime, timedelta

from soda.common.exception_helper import get_exception_stacktrace
from soda.sampler.db_sample import DbSample
from soda.sampler.sample_context import SampleContext
from soda.sampler.sampler import Sampler


class Query:
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        table: Table = None,
        partition: Partition = None,
        column: Column = None,
        unqualified_query_name: str = None,
        sql: str | None = None,
        sample_name: str = "failed_rows",
    ):
        self.logs = data_source_scan.scan._logs
        self.data_source_scan = data_source_scan
        self.query_name: str = Query.build_query_name(
            data_source_scan, table, partition, column, unqualified_query_name
        )
        self.sample_name = sample_name
        self.table: Table | None = table
        self.partition: Partition | None = partition
        self.column: Column | None = column

        # The SQL query that is used _fetchone or _fetchall or _store
        # This field can also be initialized in the execute method before any of _fetchone,
        # _fetchall or _store are called
        self.sql: str = sql

        # Following fields are initialized in execute method
        self.description: tuple | None = None
        self.row: tuple | None = None
        self.rows: list[tuple] | None = None
        self.sample_ref: SampleRef | None = None
        self.exception: BaseException | None = None
        self.duration: timedelta | None = None

    def get_cloud_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        return {
            "name": self.query_name,
            "dataSource": self.data_source_scan.data_source.data_source_name,
            "table": Partition.get_table_name(self.partition),
            "partition": Partition.get_partition_name(self.partition),
            "column": Column.get_partition_name(self.column),
            "sql": self.sql,
            "exception": get_exception_stacktrace(self.exception),
            "duration": self.duration,
        }

    @staticmethod
    def build_query_name(data_source_scan, table, partition, column, unqualified_query_name):
        full_query_pieces = [data_source_scan.data_source.data_source_name]
        if partition is not None and partition.partition_name is not None:
            full_query_pieces.append(f"{partition.table.table_name}[{partition.partition_name}]")
        elif table is not None:
            full_query_pieces.append(f"{table.table_name}")
        if column is not None:
            full_query_pieces.append(f"{column.column_name}")
        full_query_pieces.append(unqualified_query_name)
        return ".".join(full_query_pieces)

    def execute(self):
        """
        Execute method implementations should
          - invoke either self.fetchone, self.fetchall or self.store
          - update the metrics with value and optionally other diagnostic information
        """
        self.fetchall()

    def fetchone(self):
        """
        DataSource query execution exceptions will be caught and result in the
        self.exception being populated.
        """
        self.__append_to_scan()
        start = datetime.now()
        data_source = self.data_source_scan.data_source
        try:
            cursor = data_source.connection.cursor()
            try:
                self.logs.debug(f"Query {self.query_name}:\n{self.sql}")
                cursor.execute(self.sql)
                self.row = cursor.fetchone()
                self.description = cursor.description
            finally:
                cursor.close()
        except BaseException as e:
            self.exception = e
            self.logs.error(
                f"Query execution error in {self.query_name}: {e}\n{self.sql}",
                e,
            )
            data_source.query_failed(e)
        finally:
            self.duration = datetime.now() - start

    def fetchall(self):
        """
        DataSource query execution exceptions will be caught and result in the
        self.exception being populated.
        """
        self.__append_to_scan()
        start = datetime.now()
        data_source = self.data_source_scan.data_source
        try:
            cursor = data_source.connection.cursor()
            try:
                self.logs.debug(f"Query {self.query_name}:\n{self.sql}")
                cursor.execute(self.sql)
                self.rows = cursor.fetchall()
                self.description = cursor.description
            finally:
                cursor.close()
        except BaseException as e:
            self.exception = e
            self.logs.error(f"Query error: {self.query_name}: {e}\n{self.sql}", exception=e)
            data_source.query_failed(e)
        finally:
            self.duration = datetime.now() - start

    def store(self):
        """
        DataSource query execution exceptions will be caught and result in the
        self.exception being populated.
        """
        self.__append_to_scan()
        sampler: Sampler = self.data_source_scan.scan._configuration.sampler
        if sampler:
            start = datetime.now()
            data_source = self.data_source_scan.data_source
            try:
                cursor = data_source.connection.cursor()
                try:
                    self.logs.debug(f"Query {self.query_name}:\n{self.sql}")
                    cursor.execute(self.sql)
                    self.description = cursor.description

                    db_sample = DbSample(cursor, self.data_source_scan.data_source)

                    sample_context = SampleContext(
                        sample=db_sample,
                        sample_name=self.sample_name,
                        query=self.sql,
                        data_source=self.data_source_scan.data_source,
                        partition=self.partition,
                        column=self.column,
                        scan=self.data_source_scan.scan,
                        logs=self.data_source_scan.scan._logs,
                    )

                    self.sample_ref = sampler.store_sample(sample_context)
                finally:
                    cursor.close()
            except BaseException as e:
                self.exception = e
                self.logs.error(f"Query error: {self.query_name}: {e}\n{self.sql}", exception=e)
                data_source.query_failed(e)
            finally:
                self.duration = datetime.now() - start
        else:
            self.logs.debug(f"Not executing query {self.query_name} because no sampler is configured.")

    def __append_to_scan(self):
        scan = self.data_source_scan.scan
        self.index = len(scan._queries)
        scan._queries.append(self)



from typing import List

from soda.profiling.discover_table_result_table import DiscoverTablesResultTable
from soda.sodacl.discover_tables_cfg import DiscoverTablesCfg


class DiscoverTablesResult:
    def __init__(self, discover_tables_cfg: DiscoverTablesCfg):
        self.discover_tables_cfg: DiscoverTablesCfg = discover_tables_cfg
        self.tables: List[DiscoverTablesResultTable] = []

    def create_table(self, table_name: str, data_source_name: str, row_count: int) -> DiscoverTablesResultTable:
        table = DiscoverTablesResultTable(table_name, data_source_name, row_count)
        self.tables.append(table)
        return table



from __future__ import annotations

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location
from soda.sodacl.threshold_cfg import ThresholdCfg


class DistributionCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: str | None,
        location: Location,
        name: str | None,
        column_name: str,
        distribution_name: str,
        filter: str,
        reference_file_path: str,
        fail_threshold_cfg: ThresholdCfg | None,
        warn_threshold_cfg: ThresholdCfg | None,
    ):
        super().__init__(
            source_header=source_header,
            source_line=source_line,
            source_configurations=source_configurations,
            location=location,
            name=name,
        )
        self.column_name = column_name
        self.distribution_name = distribution_name
        self.filter = filter
        self.reference_file_path = reference_file_path
        self.fail_threshold_cfg = fail_threshold_cfg
        self.warn_threshold_cfg = warn_threshold_cfg

    def get_column_name(self) -> str | None:
        return self.column_name



from __future__ import annotations

from soda.sampler.sample_schema import SampleSchema


class SampleRef:
    def __init__(
        self,
        # Sample display name for UIs
        name: str,
        schema: SampleSchema,
        total_row_count: int,
        stored_row_count: int,
        type: str,
        soda_cloud_file_id: str | None = None,
        message: str | None = None,
        link: str | None = None,
    ):
        self.name: str = name
        self.schema = schema
        self.total_row_count: int = total_row_count
        self.stored_row_count: int = stored_row_count
        self.type: str = type
        self.soda_cloud_file_id: str | None = soda_cloud_file_id
        self.message: str | None = message
        self.link: str | None = link

    def __str__(self) -> str:
        sample_dimension = f"{len(self.schema.columns)}x({self.stored_row_count}/{self.total_row_count})"
        return " ".join(
            [
                e
                for e in [self.type, self.soda_cloud_file_id, self.message, self.link, sample_dimension]
                if e is not None
            ]
        )

    def get_cloud_diagnostics_dict(self):
        column_dicts = [column.get_cloud_dict() for column in self.schema.columns]
        sample_ref_dict = {
            "columns": column_dicts,
            "totalRowCount": self.total_row_count,
            "storedRowCount": self.stored_row_count,
        }
        if self.soda_cloud_file_id:
            sample_ref_dict["reference"] = {"type": "sodaCloudStorage", "fileId": self.soda_cloud_file_id}

        if self.message:
            sample_ref_dict["reference"] = {"type": "externalStorage", "message": self.message, "link": self.link}
        return sample_ref_dict



from typing import Dict, List

from soda.common.exceptions import DataSourceError
from soda.execution.data_source import DataSource


class DataSourceManager:
    """
    Caches data_sources and manages connections for data_sources
    """

    def __init__(self, logs: "Logs", configuration: "Configuration"):
        self.logs = logs
        self.configuration = configuration
        self.connection_properties_by_name: Dict[str, dict] = configuration.connection_properties_by_name
        self.data_source_properties_by_name: Dict[str, dict] = configuration.data_source_properties_by_name
        self.connections: Dict[str, object] = {}
        self.data_sources: Dict[str, DataSource] = {}

    def get_data_source_names(self) -> List[str]:
        return list(self.data_source_properties_by_name.keys())

    def get_data_source(self, data_source_name: str) -> DataSource:
        """
        Returns a data_source.
        """
        data_source = self.data_sources.get(data_source_name)
        # No error generation needed as it should be checked during the parsing
        if not data_source:
            data_source_properties = self.data_source_properties_by_name.get(data_source_name)
            if data_source_properties:
                connection_name = data_source_properties.get("connection")
                if connection_name:
                    connection_properties = self.connection_properties_by_name.get(connection_name)
                    if connection_properties:
                        connection_type = data_source_properties.get("type")
                        if connection_type:
                            data_source = DataSource.create(
                                self.logs,
                                data_source_name,
                                connection_type,
                                data_source_properties,
                                connection_properties,
                            )
                            if data_source:
                                try:
                                    data_source.connect(connection_properties)
                                    self.data_sources[data_source_name] = data_source
                                except BaseException as e:
                                    self.logs.error(
                                        f'Could not connect to data source "{data_source_name}": {e}', exception=e
                                    )
                                    data_source = None
                        else:
                            self.logs.error(f'Data source "{data_source_name}" does not have a type')
            else:
                raise DataSourceError(f"Data source '{data_source_name}' not present in the configuration.")

        return data_source

    def connect(self, data_source: DataSource) -> object:
        if not data_source.connection:
            connection_name = data_source.data_source_properties.get("connection")
            data_source.connection = self._get_connection(connection_name, data_source)
        return data_source.connection

    def close_all_connections(self):
        for connection_name, connection in self.connections.items():
            try:
                connection.close()
            except BaseException as e:
                self.logs.error(f"Could not close connection {connection_name}: {e}", exception=e)

    def _get_connection(self, connection_name: str, data_source: DataSource) -> object:
        """
        Returns a connection.
        """
        connection = self.connections.get(connection_name)
        if connection is None:
            connection_properties = self.connection_properties_by_name.get(connection_name)
            return data_source.connect(connection_properties)
        return connection



from soda.sampler.sample_context import SampleContext
from soda.sampler.sample_ref import SampleRef
from soda.sampler.sampler import Sampler


class SodaCloudSampler(Sampler):
    def store_sample(self, sample_context: SampleContext) -> SampleRef:
        sample_rows = sample_context.sample.get_rows()
        row_count = len(sample_rows)
        column_count = len(sample_rows[0]) if row_count > 0 else 0

        scan = sample_context.scan
        soda_cloud = scan._configuration.soda_cloud

        soda_cloud_file_id = soda_cloud.upload_sample(
            scan=scan, sample_rows=sample_rows, sample_file_name=sample_context.get_sample_file_name()
        )

        return SampleRef(
            name=sample_context.sample_name,
            schema=sample_context.sample.get_schema(),
            total_row_count=row_count,
            stored_row_count=row_count,
            type="soda_cloud",
            soda_cloud_file_id=soda_cloud_file_id,
        )



from typing import List

from soda.execution.query import Query


class AggregationQuery(Query):
    def __init__(self, partition: "Partition", aggregation_query_index: int):
        super().__init__(
            data_source_scan=partition.data_source_scan,
            table=partition.table,
            partition=partition,
            unqualified_query_name=f"aggregation[{aggregation_query_index}]",
        )
        from soda.execution.query_metric import QueryMetric

        self.select_expressions: List[str] = []
        self.metrics: List[QueryMetric] = []

    def add_metric(self, sql_expression: str, metric: "Metric"):
        self.select_expressions.append(sql_expression)
        self.metrics.append(metric)

    def execute(self):
        scan = self.data_source_scan.scan
        select_expression_sql = f",\n  ".join(self.select_expressions)
        self.sql = f"SELECT \n" f"  {select_expression_sql} \n" f"FROM {self.partition.table.prefixed_table_name}"

        partition_filter = self.partition.sql_partition_filter
        if partition_filter:
            resolved_filter = scan._jinja_resolve(definition=partition_filter)
            self.sql += f"\nWHERE {resolved_filter}"

        self.fetchone()
        if self.row:
            for i in range(0, len(self.row)):
                metric = self.metrics[i]
                fetched_value = self.row[i]
                metric.set_value(fetched_value)

                sample_query = metric.create_failed_rows_sample_query()
                if sample_query:
                    sample_query.execute()



from soda.execution.query import Query


class UserDefinedFailedRowsExpressionQuery(Query):
    def __init__(self, data_source_scan: "DataSourceScan", check_name: str, sql: str):
        super().__init__(
            data_source_scan=data_source_scan,
            unqualified_query_name=f"user_defined_failed_rows_expression_query[{check_name}]",
        )
        self.sql: str = sql

    def execute(self):
        self.store()



from soda.execution.query import Query


class DuplicatesQuery(Query):
    def __init__(self, partition: "Partition", metric: "Metric"):
        super().__init__(
            data_source_scan=partition.data_source_scan,
            table=partition.table,
            partition=partition,
            column=metric.column,
            unqualified_query_name=f"duplicate_count",
        )
        self.metric = metric

        values_filter_clauses = [f"{column_name} IS NOT NULL" for column_name in self.metric.metric_args]
        partition_filter = self.partition.sql_partition_filter
        if partition_filter:
            scan = self.data_source_scan.scan
            location = self.metric._checks[0].location if self.metric._checks else None
            resolved_partition_filter = scan._jinja_resolve(definition=partition_filter, location=location)
            values_filter_clauses.append(resolved_partition_filter)

        values_filter = " \n  AND ".join(values_filter_clauses)

        column_names = ", ".join(self.metric.metric_args)

        self.sql = (
            f"WITH frequencies AS (\n"
            f"  SELECT {column_names}, COUNT(*) AS frequency \n"
            f"  FROM {self.partition.table.prefixed_table_name} \n"
            f"  WHERE {values_filter} \n"
            f"  GROUP BY {column_names}) \n"
            f"SELECT * \n"
            f"FROM frequencies \n"
            f"WHERE frequency > 1;"
        )

    def execute(self):
        self.store()
        if self.sample_ref:
            values_having_duplicates = self.sample_ref.total_row_count
            self.metric.set_value(values_having_duplicates)
            self.metric.failed_rows_sample_ref = self.sample_ref



from __future__ import annotations

from abc import ABC

from soda.common.undefined_instance import undefined
from soda.execution.identity import Identity
from soda.execution.query import Query
from soda.sampler.sample_ref import SampleRef


class Metric(ABC):
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        partition: Partition | None,
        column: Column | None,
        name: str,
        check: Check,
        # schedule name, data source name, table name, partition name, column name and name are
        # already added to the identity.  All other metric configurations that are part of the
        # identity should be provided in this identity_hash_parts list.
        # Only a combination of list, dict, str, int, float is allowed.
        identity_parts: list,
    ):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        # Only used in the string representation of metrics.  See __str__(self) below
        self.name: str = name
        # identity will be used to resolve the same metric and bind it to different checks.
        # Unique within an organisation, but cannot be used as global id in Soda Cloud.
        if hasattr(check.check_cfg, "is_automated_monitoring"):
            is_automated_monitoring = check.check_cfg.is_automated_monitoring
        else:
            is_automated_monitoring = False
        self.identity: str = Identity.create_identity(
            "metric", data_source_scan, partition, column, name, is_automated_monitoring, identity_parts
        )
        self.data_source_scan = data_source_scan
        self.partition: Partition = partition
        self.column: Column | None = column

        self.checks: set[Check] = {check}

        self.value: object = undefined
        self.queries: list[Query] = []
        self.formula_values: dict[str, object] = None
        self.failed_rows_sample_ref: SampleRef | None = None

    def __eq__(self, other: Metric) -> bool:
        if self is other:
            return True
        if type(self) != type(other):
            return False
        return self.identity == other.identity

    def __hash__(self) -> int:
        return hash(self.identity)

    def __str__(self):
        return self.name

    def merge_checks(self, other_metric: Metric):
        if other_metric.checks is not None:
            self.checks.update(other_metric.checks)

    def set_value(self, value):
        self.value = value

    def ensure_query(self):
        """
        Every query should be added to either to the partition.queries or to data_source_scan.queries
        """

    def get_cloud_dict(self):
        pass

        return {
            "identity": self.identity,
            "metricName": self.name,
            "value": self.value,
            # TODO: re-enable once backend supports these properties.
            # "dataSourceName": self.data_source_scan.data_source.data_source_name,
            # "tableName": Partition.get_table_name(self.partition),
            # "filterName": Partition.get_partition_name(self.partition),
            # "columnName": Column.get_partition_name(self.column),
        }



import json
from typing import Dict, List, Sequence

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult


class MemorySpanExporter(SpanExporter):
    """Implementation of :class:`SpanExporter` that saves spans in memory.

    This class can be used for diagnostic purposes, multi-threaded scenarios etc.
    """

    __instance = None
    __spans = []

    @staticmethod
    def get_instance():
        if MemorySpanExporter.__instance is None:
            MemorySpanExporter()
        return MemorySpanExporter.__instance

    def __init__(self):
        if MemorySpanExporter.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            MemorySpanExporter.__instance = self

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for span in spans:
            self.__spans.append(span)
        return SpanExportResult.SUCCESS

    def reset(self):
        self.__spans = []

    @property
    def spans(self) -> List[ReadableSpan]:
        return self.__spans

    @property
    def span_dicts(self) -> List[Dict]:
        return [json.loads(span.to_json()) for span in self.spans]



from __future__ import annotations

from typing import TYPE_CHECKING

from soda.execution.query import Query
from soda.profiling.discover_tables_result import DiscoverTablesResult
from soda.sodacl.discover_tables_cfg import DiscoverTablesCfg

if TYPE_CHECKING:
    from soda.execution.data_source_scan import DataSourceScan


QUERY_PREFIX = "discover_tables: "


class DiscoverTablesRun:
    def __init__(self, data_source_scan: DataSourceScan, discover_tables_cfg: DiscoverTablesCfg):

        self.data_source_scan = data_source_scan
        self.soda_cloud = data_source_scan.scan._configuration.soda_cloud
        self.data_source = data_source_scan.data_source
        self.discover_tables_cfg: DiscoverTablesCfg = discover_tables_cfg
        self.logs = self.data_source_scan.scan._logs

    def run(self) -> DiscoverTablesResult:
        discover_tables_result: DiscoverTablesResult = DiscoverTablesResult(self.discover_tables_cfg)
        self.logs.info(f"Running discover tables for data source: {self.data_source.data_source_name}")

        # row_counts is a dict that maps table names to row counts.
        row_counts_by_table_name: dict[str, int] = self.data_source.get_row_counts_all_tables(
            include_tables=self.discover_tables_cfg.include_tables,
            exclude_tables=self.discover_tables_cfg.exclude_tables,
            query_name=f"{QUERY_PREFIX} get tables and row counts",
        )
        for table_name in row_counts_by_table_name:
            self.logs.debug(f"Discovering columns for {table_name}")
            measured_row_count = row_counts_by_table_name[table_name]
            discover_tables_result_table = discover_tables_result.create_table(
                table_name, self.data_source.data_source_name, measured_row_count
            )

            # get columns & metadata for current table
            columns_metadata_sql = self.data_source.sql_to_get_column_metadata_for_table(table_name)
            columns_metadata_query = Query(
                data_source_scan=self.data_source_scan,
                unqualified_query_name=f"{QUERY_PREFIX} get col metadata for table: {table_name}",
                sql=columns_metadata_sql,
            )
            columns_metadata_query.execute()
            assert columns_metadata_query.rows, f"No metadata was captured for table: {table_name}"
            columns_metadata_result = {column[0]: column[1] for column in columns_metadata_query.rows}
            for column_name, column_type in columns_metadata_result.items():
                _ = discover_tables_result_table.create_column(column_name, column_type)

        if not discover_tables_result.tables:
            self.logs.error(f"Discover tables for data source: {self.data_source.data_source_name} failed")
        return discover_tables_result



from typing import List

from soda.sodacl.name_filter import NameFilter


class ColumnsetCfg:
    def __init__(self, columnset_name: str):
        self.columnset_name: str = columnset_name
        self.includes: List[NameFilter] = []
        self.excludes: List[NameFilter] = []



from __future__ import annotations

from soda.sodacl.column_checks_cfg import ColumnChecksCfg


class ColumnMetrics:
    def __init__(self, partition: Partition, column_name: str):
        from soda.execution.metric import Metric
        from soda.execution.partition import Partition
        from soda.sodacl.column_checks_cfg import ColumnChecksCfg

        self.data_source_scan = partition.data_source_scan
        self.partition: Partition = partition
        self.column = partition.table.get_or_create_column(column_name)
        self.column_checks_cfg: ColumnChecksCfg = None
        self.metrics: list[Metric] = []

    def set_column_check_cfg(self, column_checks_cfg: ColumnChecksCfg):
        self.column_checks_cfg = column_checks_cfg

    def add_check_cfg(self, check_cfg):
        if self.column_checks_cfg is None:
            self.column_checks_cfg = ColumnChecksCfg(self.column.column_name)
        self.column_checks_cfg.add_check_cfg(check_cfg)
